# -*- coding: utf-8 -*-

# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TypeVar, Optional, Any, Union, TYPE_CHECKING

from iconservice.icon_constant import IconScoreContextType, Revision
from .context.context import ContextContainer
from ..base.address import Address
from ..base.exception import InvalidParamsException, InvalidContainerAccessException
from ..utils import int_to_bytes, bytes_to_int

if TYPE_CHECKING:
    from ..database.db import IconScoreDatabase, IconScoreSubDatabase

K = TypeVar('K', int, str, Address, bytes)
V = TypeVar('V', int, str, Address, bytes, bool)

ARRAY_DB_ID = b'\x00'
DICT_DB_ID = b'\x01'
VAR_DB_ID = b'\x02'


def get_encoded_key(key: V) -> bytes:
    return ContainerUtil.encode_key(key)


class ContainerUtil(object):

    @classmethod
    def create_db_prefix(cls, container_cls: type, var_key: K) -> bytes:
        """Create a prefix used
        as a parameter of IconScoreDatabase.get_sub_db()

        :param container_cls: ArrayDB, DictDB, VarDB
        :param var_key:
        :return:
        """
        if container_cls == ArrayDB:
            container_id = ARRAY_DB_ID
        elif container_cls == DictDB:
            container_id = DICT_DB_ID
        else:
            raise InvalidParamsException(f'Unsupported container class: {container_cls}')

        encoded_key: bytes = get_encoded_key(var_key)
        return b'|'.join([container_id, encoded_key])

    @classmethod
    def encode_key(cls, key: K) -> bytes:
        """Create a key passed to IconScoreDatabase

        :param key:
        :return:
        """
        if key is None:
            raise InvalidParamsException('key is None')

        if isinstance(key, int):
            bytes_key = int_to_bytes(key)
        elif isinstance(key, str):
            bytes_key = key.encode('utf-8')
        elif isinstance(key, Address):
            bytes_key = key.to_bytes()
        elif isinstance(key, bytes):
            bytes_key = key
        else:
            raise InvalidParamsException(f'Unsupported key type: {type(key)}')
        return bytes_key

    @classmethod
    def encode_value(cls, value: V) -> bytes:
        if isinstance(value, int):
            byte_value = int_to_bytes(value)
        elif isinstance(value, str):
            byte_value = value.encode('utf-8')
        elif isinstance(value, Address):
            byte_value = value.to_bytes()
        elif isinstance(value, bool):
            byte_value = int_to_bytes(int(value))
        elif isinstance(value, bytes):
            byte_value = value
        else:
            raise InvalidParamsException(f'Unsupported value type: {type(value)}')
        return byte_value

    @classmethod
    def decode_object(cls, value: bytes, value_type: type) -> Optional[Union[K, V]]:
        if value is None:
            return get_default_value(value_type)

        obj_value = None
        if value_type == int:
            obj_value = bytes_to_int(value)
        elif value_type == str:
            obj_value = value.decode()
        elif value_type == Address:
            obj_value = Address.from_bytes(value)
        if value_type == bool:
            obj_value = bool(bytes_to_int(value))
        elif value_type == bytes:
            obj_value = value
        return obj_value

    @classmethod
    def remove_prefix_from_iters(cls, iter_items: iter) -> iter:
        return ((cls.__remove_prefix_from_key(key), value) for key, value in iter_items)

    @classmethod
    def __remove_prefix_from_key(cls, key_from_bytes: bytes) -> bytes:
        return key_from_bytes[:-1]

    @classmethod
    def put_to_db(cls, db: 'IconScoreDatabase', db_key: str, container: iter) -> None:
        sub_db = db.get_sub_db(cls.encode_key(db_key))
        if isinstance(container, dict):
            cls.__put_to_db_internal(sub_db, container.items())
        elif isinstance(container, (list, set, tuple)):
            cls.__put_to_db_internal(sub_db, enumerate(container))

    @classmethod
    def get_from_db(cls, db: 'IconScoreDatabase', db_key: str, *args, value_type: type) -> Optional[K]:
        sub_db = db.get_sub_db(cls.encode_key(db_key))
        *args, last_arg = args
        for arg in args:
            sub_db = sub_db.get_sub_db(cls.encode_key(arg))

        byte_key = sub_db.get(cls.encode_key(last_arg))
        if byte_key is None:
            return get_default_value(value_type)
        return cls.decode_object(byte_key, value_type)

    @classmethod
    def __put_to_db_internal(cls, db: Union['IconScoreDatabase', 'IconScoreSubDatabase'], iters: iter) -> None:
        for key, value in iters:
            sub_db = db.get_sub_db(cls.encode_key(key))
            if isinstance(value, dict):
                cls.__put_to_db_internal(sub_db, value.items())
            elif isinstance(value, (list, set, tuple)):
                cls.__put_to_db_internal(sub_db, enumerate(value))
            else:
                db_key = cls.encode_key(key)
                db_value = cls.encode_value(value)
                db.put(db_key, db_value)


class DictDB(object):
    """
    Utility classes wrapping the state DB.
    DictDB behaves more like python dict.
    DictDB does not maintain order.

    :K: [int, str, Address, bytes]
    :V: [int, str, Address, bytes, bool]
    """

    def __init__(self,
                 var_key: K,
                 db: Union['IconScoreDatabase',
                           'IconScoreSubDatabase'],
                 value_type: type,
                 depth: int = 1) -> None:
        prefix: bytes = ContainerUtil.create_db_prefix(type(self), var_key)
        self._db = db.get_sub_db(prefix)

        self.__value_type = value_type
        self.__depth = depth

    def remove(self, key: K) -> None:
        """
        Removes the value of given key

        :param key:
        """
        self.__remove(key)

    def __setitem__(self, key: K, value: V) -> None:
        if self.__depth != 1:
            raise InvalidContainerAccessException('DictDB depth mismatch')

        encoded_key: bytes = get_encoded_key(key)
        encoded_value: bytes = ContainerUtil.encode_value(value)

        self._db.put(encoded_key, encoded_value)

    def __getitem__(self, key: K) -> Any:
        if self.__depth == 1:
            encoded_key: bytes = get_encoded_key(key)
            return ContainerUtil.decode_object(self._db.get(encoded_key), self.__value_type)
        else:
            return DictDB(key, self._db, self.__value_type, self.__depth - 1)

    def __delitem__(self, key: K):
        self.__remove(key)

    def __contains__(self, key: K):
        # Plyvel doesn't allow setting None value in the DB.
        # so there is no case of returning None value if the key exists.
        value = self._db.get(get_encoded_key(key))
        return value is not None

    def __remove(self, key: K) -> None:
        if self.__depth != 1:
            raise InvalidContainerAccessException('DictDB depth mismatch')
        self._db.delete(get_encoded_key(key))

    def __iter__(self):
        raise InvalidContainerAccessException("Iteration not supported in DictDB")


class ArrayDB(object):
    """
    Utility classes wrapping the state DB.
    ArrayDB supports length and iterator, maintains order.

    :K: [int, str, Address, bytes]
    :V: [int, str, Address, bytes, bool]
    """
    __SIZE = 'size'
    __SIZE_BYTE_KEY = get_encoded_key(__SIZE)

    def __init__(self, var_key: K, db: 'IconScoreDatabase', value_type: type) -> None:
        prefix: bytes = ContainerUtil.create_db_prefix(type(self), var_key)
        self._db = db.get_sub_db(prefix)
        self.__value_type = value_type
        self.__legacy_size = self.__get_size_from_db()

    def put(self, value: V) -> None:
        """
        Puts the value at the end of array

        :param value: value to add
        """
        size: int = self.__get_size()
        self.__put(size, value)
        self.__set_size(size + 1)

    def pop(self) -> Optional[V]:
        """
        Gets and removes last added value

        :return: last added value
        """
        size: int = self.__get_size()
        if size == 0:
            return None

        index = size - 1
        last_val = self[index]
        self._db.delete(get_encoded_key(index))
        self.__set_size(index)
        return last_val

    def get(self, index: int = 0) -> V:
        """
        Gets the value at index

        :param index: index
        :return: value at the index
        """
        return self[index]

    def __get_size(self) -> int:
        if self.__is_defective_revision():
            return self.__legacy_size
        else:
            return self.__get_size_from_db()

    def __get_size_from_db(self) -> int:
        return ContainerUtil.decode_object(self._db.get(self.__SIZE_BYTE_KEY), int)

    def __set_size(self, size: int) -> None:
        self.__legacy_size = size
        byte_value = ContainerUtil.encode_value(size)
        self._db.put(self.__SIZE_BYTE_KEY, byte_value)

    def __put(self, index: int, value: V) -> None:
        byte_value = ContainerUtil.encode_value(value)
        self._db.put(get_encoded_key(index), byte_value)

    def __iter__(self):
        return self._get_generator(self._db, self.__get_size(), self.__value_type)

    def __len__(self):
        return self.__get_size()

    def __setitem__(self, index: int, value: V) -> None:
        if not isinstance(index, int):
            raise InvalidParamsException('Invalid index type: not an integer')

        size: int = self.__get_size()

        # Negative index means that you count from the right instead of the left.
        if index < 0:
            index += size

        if 0 <= index < size:
            self.__put(index, value)
        else:
            raise InvalidParamsException('ArrayDB out of index')

    def __getitem__(self, index: int) -> V:
        return self._get(self._db, self.__get_size(), index, self.__value_type)

    def __contains__(self, item: V):
        for e in self:
            if e == item:
                return True
        return False

    @classmethod
    def __is_defective_revision(cls):
        context = ContextContainer._get_context()
        revision = context.revision
        return context.type == IconScoreContextType.INVOKE and revision < Revision.THREE.value

    @classmethod
    def _get(cls, db: Union['IconScoreDatabase', 'IconScoreSubDatabase'], size: int, index: int, value_type: type) -> V:
        if not isinstance(index, int):
            raise InvalidParamsException('Invalid index type: not an integer')

        # Negative index means that you count from the right instead of the left.
        if index < 0:
            index += size

        if 0 <= index < size:
            key: bytes = get_encoded_key(index)
            return ContainerUtil.decode_object(db.get(key), value_type)

        raise InvalidParamsException('ArrayDB out of index')

    @classmethod
    def _get_generator(cls, db: Union['IconScoreDatabase', 'IconScoreSubDatabase'], size: int, value_type: type):
        for index in range(size):
            yield cls._get(db, size, index, value_type)


class VarDB(object):
    """
    Utility classes wrapping the state DB.
    VarDB can be used to store simple key-value state.

    :K: [int, str, Address, bytes]
    :V: [int, str, Address, bytes, bool]
    """

    def __init__(self, var_key: K, db: 'IconScoreDatabase', value_type: type) -> None:
        # Use var_key as a db prefix in the case of VarDB
        self._db = db.get_sub_db(VAR_DB_ID)
        self.__var_byte_key = get_encoded_key(var_key)
        self.__value_type = value_type

    def set(self, value: V) -> None:
        """
        Sets the value

        :param value: a value to be set
        """
        byte_value = ContainerUtil.encode_value(value)
        self._db.put(self.__var_byte_key, byte_value)

    def get(self) -> Optional[V]:
        """
        Gets the value

        :return: value of the var db
        """
        return ContainerUtil.decode_object(self._db.get(self.__var_byte_key), self.__value_type)

    def remove(self) -> None:
        """
        Deletes the value
        """
        self._db.delete(self.__var_byte_key)


def get_default_value(value_type: type) -> Any:
    if value_type == int:
        return 0
    elif value_type == str:
        return ""
    elif value_type == bool:
        return False
    return None



# -*- coding: utf-8 -*-

# Copyright 2019 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING, Optional, List

from .base_part import BasePart, BasePartState
from ..base.exception import InvalidParamsException
from ..icon_constant import Revision
from ..utils.msgpack_for_db import MsgPackForDB

if TYPE_CHECKING:
    from ..base.address import Address


class StakePart(BasePart):
    _VERSION = 0
    PREFIX = b"aos|"

    def __init__(self,
                 stake: int = 0,
                 unstake: int = 0,
                 unstake_block_height: int = 0,
                 unstakes_info: Optional[List] = None):
        super().__init__()

        self._stake: int = stake
        self._unstake: int = unstake
        self._unstake_block_height: int = unstake_block_height
        self._unstakes_info: List[List[int, int]] = unstakes_info if unstakes_info else []

    def __str__(self):
        return f"stake={self._stake}, " \
               f"unstake={self._unstake}, " \
               f"unstake_bh={self._unstake_block_height}"

    @staticmethod
    def make_key(address: 'Address') -> bytes:
        return StakePart.PREFIX + address.to_bytes_including_prefix()

    @property
    def stake(self) -> int:
        assert self.is_set(BasePartState.COMPLETE)
        return self._stake

    @property
    def voting_weight(self) -> int:
        assert self.is_set(BasePartState.COMPLETE)
        return self._stake

    @property
    def unstake(self) -> int:
        assert self.is_set(BasePartState.COMPLETE)
        return self._unstake

    @property
    def unstake_block_height(self) -> int:
        assert self.is_set(BasePartState.COMPLETE)
        return self._unstake_block_height

    @property
    def total_stake(self) -> int:
        assert self.is_set(BasePartState.COMPLETE)
        return self._stake + self.total_unstake

    @property
    def unstakes_info(self) -> List:
        assert self.is_set(BasePartState.COMPLETE)
        return self._unstakes_info

    @property
    def total_unstake(self) -> int:
        assert self.is_set(BasePartState.COMPLETE)
        return self._total_unstake()

    def _total_unstake(self) -> int:
        if self._unstakes_info:
            return sum(map(lambda unstakes_info: unstakes_info[0], self._unstakes_info)) + self._unstake
        return self._unstake

    def add_stake(self, value: int):
        assert self.is_set(BasePartState.COMPLETE)

        if value <= 0:
            raise InvalidParamsException("Failed to stake: value <= 0")

        self._stake += value
        self.set_dirty(True)

    def set_unstake(self, block_height: int, value: int):
        assert self.is_set(BasePartState.COMPLETE)

        if self.total_stake < value:
            raise InvalidParamsException(f'Failed to unstake: stake_amount({self._stake}) < value({value})')

        # FIXME: Consider the case when value is 0
        self._stake = self.total_stake - value
        self._unstake = value
        self._unstake_block_height: int = block_height

        self.set_dirty(True)

    def set_unstakes_info(self, block_height: int, new_total_unstake: int, slot_max: int):
        total_stake = self.total_stake

        if new_total_unstake < self.total_unstake:
            self.withdraw_unstake(self.total_unstake - new_total_unstake)

        elif self.total_unstake < new_total_unstake:
            increment_unstake = new_total_unstake - self.total_unstake
            if len(self._unstakes_info) == slot_max:
                old_value_pair = self._unstakes_info.pop()
                increment_unstake += old_value_pair[0]
                unstake_block_height = max(old_value_pair[1], block_height)
                self._unstakes_info.append([increment_unstake, unstake_block_height])
            else:
                new_value_index = self.get_unstake_slot_index(block_height)
                self._unstakes_info.insert(new_value_index, [increment_unstake, block_height])

        self._stake = total_stake - new_total_unstake
        self.set_dirty(True)

    def withdraw_unstake(self, amount: int):
        unstakes_length = len(self._unstakes_info)
        accumulated_unstake = 0
        total_unstake = self.total_unstake
        new_total_unstake = total_unstake - amount
        for index in range(unstakes_length):
            accumulated_unstake += self.unstakes_info[index][0]
            if new_total_unstake > accumulated_unstake:
                continue
            elif new_total_unstake == accumulated_unstake:
                self._unstakes_info = self.unstakes_info[:index + 1]
                return
            elif new_total_unstake < accumulated_unstake:
                self._unstakes_info = self.unstakes_info[:index + 1]
                old_value_pair = self._unstakes_info.pop()
                new_value_pair = \
                    [(new_total_unstake - accumulated_unstake + old_value_pair[0]), old_value_pair[1]]
                self._unstakes_info.append(new_value_pair)
                return

    def reset_unstake(self):
        assert self.is_set(BasePartState.COMPLETE)

        self._stake = self.total_stake
        self._unstake = 0
        self._unstake_block_height: int = 0
        self._unstakes_info = []

        self.set_dirty(True)

    def normalize(self, block_height: int, revision: int) -> int:
        unstake: int = 0
        state: 'BasePartState' = BasePartState.COMPLETE

        if revision >= Revision.MULTIPLE_UNSTAKE.value:
            if self._unstake_block_height:
                self._unstakes_info.append([self._unstake, self._unstake_block_height])
                self._unstake = 0
                self._unstake_block_height = 0

            size = len(self._unstakes_info)
            for i in range(size):
                info = self._unstakes_info[0]
                if info[1] >= block_height:
                    if i > 0:
                        state |= BasePartState.DIRTY
                    break

                # Remove unstatke_info of which lock period is already expired
                self._unstakes_info.pop(0)
                unstake += info[0]

        else:
            if 0 < self._unstake_block_height < block_height:
                unstake: int = self._unstake
                self._unstake = 0
                self._unstake_block_height: int = 0

                state |= BasePartState.DIRTY

        self.toggle_state(state, True)

        return unstake

    @staticmethod
    def from_bytes(buf: bytes) -> 'StakePart':
        """Create Account of Stake object from bytes data

        :param buf: (bytes) bytes data including Account of Stake information
        :return: (AccountOfStake) AccountOfStake object
        """

        data: list = MsgPackForDB.loads(buf)
        version = data[0]

        if version == 0:
            data.append(None)

        return StakePart(stake=data[1], unstake=data[2], unstake_block_height=data[3], unstakes_info=data[4])

    def to_bytes(self, revision: int) -> bytes:
        """Convert Account of Stake object to bytes

        :return: data including information of StakePart object
        """

        assert self.is_set(BasePartState.COMPLETE)

        if revision >= Revision.MULTIPLE_UNSTAKE.value:
            version = 1
        else:
            version = 0

        data = [version,
                self._stake,
                self._unstake,
                self._unstake_block_height]

        if version >= 1:
            data.append(self._unstakes_info)

        return MsgPackForDB.dumps(data)

    def __eq__(self, other) -> bool:
        """operator == overriding

        :param other: (StakePart)
        """

        assert self.is_set(BasePartState.COMPLETE)

        return isinstance(other, StakePart) \
            and self._stake == other.stake \
            and self._unstake == other.unstake \
            and self._unstake_block_height == other.unstake_block_height \
            and self._unstakes_info == other.unstakes_info

    def __ne__(self, other) -> bool:
        """operator != overriding

        :param other: (StakePart)
        """

        assert self.is_set(BasePartState.COMPLETE)

        return not self.__eq__(other)

    def get_unstake_slot_index(self, block_height: int):
        unstakes_info = self._unstakes_info
        length = len(unstakes_info)
        for index in reversed(range(length)):
            if block_height >= unstakes_info[index][1]:
                return index + 1
        return 0



# -*- coding: utf-8 -*-
# Copyright 2020 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict
from collections.abc import MutableMapping
from inspect import (
    isfunction,
    getmembers,
    signature,
    Signature,
    Parameter,
)
from typing import Union, Mapping, List, Any, Set

from . import (
    is_base_type,
    is_struct,
    get_origin,
    get_args,
    get_annotations,
    name_to_type,
)
from . import isinstance_ex
from ..icon_score_constant import (
    CONST_SCORE_FLAG,
    ScoreFlag,
    CONST_INDEXED_ARGS_COUNT,
    CONST_CLASS_ELEMENT_METADATAS,
)
from ... import utils
from ...base.exception import (
    IllegalFormatException,
    InternalServiceErrorException,
    MethodNotFoundException,
    InvalidParamsException,
)

_VALID_SCORE_FLAG_COMBINATION = {
    ScoreFlag.EXTERNAL,
    ScoreFlag.EXTERNAL | ScoreFlag.PAYABLE,
    ScoreFlag.EXTERNAL | ScoreFlag.READONLY,
    ScoreFlag.FALLBACK | ScoreFlag.PAYABLE,
    ScoreFlag.EVENTLOG,
    ScoreFlag.INTERFACE,
}


_MAX_STRUCT_FIELDS = 16


def normalize_signature(func: callable) -> Signature:
    """Normalize signature of score methods

    1. Normalize type hint: ex) no type hint -> str
    2. Remove "self" parameter

    :param func: function attribute from class
    :return:
    """
    sig = signature(func)
    parameters = sig.parameters
    new_params = []

    normalized = False

    # CAUTION:
    # def A:
    #     def func(self):
    #         pass
    #
    #     @classmethod
    #     def cfunc(self):
    #         pass
    #
    # inspect.isfunction(A.func) == True
    # inspect.isfunction(A().func) == False
    # inspect.ismethod(A.func) == False
    # inspect.ismethod(A().func) == True
    # inspect.isfunction(A.cfunc) == False
    # inspect.ismethod(A.cfunc) == True
    # inspect.isfunction(A().cfunc) == False
    # inspect.ismethod(A().cfunc) == True
    is_regular_method: bool = isfunction(func)

    for i, k in enumerate(parameters):
        # Remove "self" parameter from signature of regular method
        if i == 0 and k == "self" and is_regular_method:
            new_param = None
        else:
            new_param = normalize_parameter(parameters[k])
            new_params.append(new_param)

        if new_param is not parameters[k]:
            normalized = True

    return_annotation = normalize_return_annotation(sig.return_annotation)
    if return_annotation is not sig.return_annotation:
        normalized = True

    if normalized:
        sig = sig.replace(parameters=new_params, return_annotation=return_annotation)

    return sig


def normalize_parameter(parameter: Parameter) -> Parameter:
    if parameter.kind != Parameter.POSITIONAL_OR_KEYWORD:
        raise IllegalFormatException(
            f"Invalid signature: name={parameter.name} kind={parameter.kind}")

    annotation = parameter.annotation

    if annotation == Parameter.empty:
        type_hint = str
    else:
        type_hint = normalize_type_hint(annotation)

    # a: int = None -> a: Union[int, None] = None
    if parameter.default is None and get_origin(type_hint) is not Union:
        type_hint = Union[type_hint, None]

    check_parameter_default_type(type_hint, parameter.default)

    if type_hint == annotation:
        # Nothing to update
        return parameter

    return parameter.replace(annotation=type_hint)


def normalize_return_annotation(return_annotation: type) -> Union[type, Signature.empty]:
    if return_annotation in (None, Signature.empty):
        return Signature.empty

    return return_annotation


def check_parameter_default_type(type_hint: type, default: Any):
    # default value type check
    if default in (Parameter.empty, None):
        return

    origin = get_origin(type_hint)

    if origin is Union:
        default_type = get_args(type_hint)[0]
    else:
        default_type = origin

    if not isinstance_ex(default, default_type):
        raise InvalidParamsException(
            f'Default value type mismatch. value={default} type={type_hint}')


def normalize_type_hint(type_hint) -> type:
    # If type hint is str, convert it to type hint
    if isinstance(type_hint, str):
        if type_hint == "Address":
            type_hint = name_to_type(type_hint)
        else:
            raise IllegalFormatException(f"Invalid type hint: {repr(type_hint)}")

    origin = get_origin(type_hint)

    if is_base_type(origin):
        return type_hint

    if is_struct(origin):
        # Check if type_hint cycling or invalid nested type hint are present
        if not is_struct_valid(origin):
            raise IllegalFormatException(f"Invalid type hint: {type_hint}")
        return type_hint

    if origin is list:
        return normalize_list_type_hint(type_hint)
    elif origin is dict:
        return normalize_dict_type_hint(type_hint)
    elif origin is Union:
        return normalize_union_type_hint(type_hint)

    raise IllegalFormatException(f"Unsupported type hint: {type_hint}")


def normalize_list_type_hint(type_hint: type) -> type:
    args = get_args(type_hint)

    if len(args) == 1:
        return List[normalize_type_hint(args[0])]

    raise IllegalFormatException(f"Invalid type hint: {type_hint}")


def normalize_dict_type_hint(type_hint: type) -> type:
    raise IllegalFormatException(f"Dict not supported: {type_hint}")

    # TODO: The below codes are commented out to prevent Dict type for being used as SCORE parameter
    # args = get_args(type_hint)
    #
    # if len(args) == 2 and args[0] is str:
    #     return Dict[str, normalize_type_hint(args[1])]
    #
    # raise IllegalFormatException(f"Invalid type hint: {type_hint}")


def normalize_union_type_hint(type_hint: type) -> type:
    args = get_args(type_hint)

    if len(args) == 2 and type(None) in args:
        arg = args[0] if args[1] is type(None) else args[1]
        if arg is not None and arg:
            return Union[normalize_type_hint(arg), None]

    raise IllegalFormatException(f"Invalid type hint: {type_hint}")


def verify_score_flag(flag: ScoreFlag, allow_payable_only: bool = True):
    """Check if score flag combination is valid

    If the flag combination is not valid, raise an exception
    """
    if flag in _VALID_SCORE_FLAG_COMBINATION:
        return

    if allow_payable_only and flag == ScoreFlag.PAYABLE:
        # Allow BH-6573883 to be synchronized for mainnet backward compatibility
        return

    raise IllegalFormatException(f"Invalid score decorator: {flag}")


class ScoreElementMetadata(object):
    def __init__(self, element: callable):
        self._signature: Signature = normalize_signature(element)
        self._element = element

    @property
    def element(self) -> callable:
        return self._element

    @property
    def name(self) -> str:
        return self._element.__name__

    @property
    def flag(self) -> ScoreFlag:
        return get_score_flag(self._element)

    @property
    def signature(self) -> Signature:
        return self._signature


class FunctionMetadata(ScoreElementMetadata):
    """Represents metadata of an exposed function in a SCORE

    """
    def __init__(self, func: callable):
        super().__init__(func)
        self._verify()

    @property
    def is_external(self) -> bool:
        return bool(self.flag & ScoreFlag.EXTERNAL)

    @property
    def is_payable(self) -> bool:
        return bool(self.flag & ScoreFlag.PAYABLE)

    @property
    def is_readonly(self) -> bool:
        return bool(self.flag & ScoreFlag.READONLY)

    @property
    def is_fallback(self) -> bool:
        return utils.is_all_flag_on(
            self.flag, ScoreFlag.FALLBACK | ScoreFlag.PAYABLE)

    def _verify(self):
        if self.is_fallback:
            self._verify_fallback_signature()

    def _verify_fallback_signature(self):
        """Verify if the signature of fallback() is valid

        fallback function must have no parameters
        """
        sig = self.signature

        if not (
                len(sig.parameters) == 0
                and sig.return_annotation in (None, Signature.empty)
        ):
            raise IllegalFormatException("Invalid fallback signature")


class EventLogMetadata(ScoreElementMetadata):
    """Represents metadata of an eventlog declared in a SCORE
    """

    def __init__(self, eventlog: callable):
        super().__init__(eventlog)

    @property
    def indexed_args_count(self) -> int:
        return getattr(self.element, CONST_INDEXED_ARGS_COUNT, 0)


class ScoreElementMetadataContainer(MutableMapping):
    """Container which has score elements like function and eventlog
    """

    def __init__(self):
        self._elements = OrderedDict()
        self._externals = 0
        self._eventlogs = 0
        self._readonly = False

    @property
    def externals(self) -> int:
        return self._externals

    @property
    def eventlogs(self) -> int:
        return self._eventlogs

    def __getitem__(self, k: str) -> ScoreElementMetadata:
        return self._elements[k]

    def __setitem__(self, k: str, v: ScoreElementMetadata) -> None:
        self._check_writable()
        self._elements[k] = v

        if isinstance(v, FunctionMetadata):
            self._externals += 1
        elif isinstance(v, EventLogMetadata):
            self._eventlogs += 1
        else:
            raise InternalServiceErrorException(f"Invalid element: {v}")

    def __iter__(self):
        for k in self._elements:
            yield k

    def __len__(self) -> int:
        return len(self._elements)

    def __delitem__(self, k: str) -> None:
        self._check_writable()

        element = self._elements[k]
        del self._elements[k]

        if is_any_score_flag_on(element, ScoreFlag.EVENTLOG):
            self._eventlogs -= 1
        else:
            self._externals -= 1

    def _check_writable(self):
        if self._readonly:
            raise InternalServiceErrorException(f"{self.__class__.__name__} not writable")

    def freeze(self):
        self._readonly = True


def create_score_element_metadatas(cls: type) -> Mapping:
    elements = ScoreElementMetadataContainer()

    for name, func in getmembers(cls, predicate=isfunction):
        if name.startswith("__"):
            continue

        # Collect the only functions with one or more of the above 4 score flags
        flag = get_score_flag(func)

        if utils.is_any_flag_on(flag, ScoreFlag.FUNC | ScoreFlag.EVENTLOG):
            verify_score_flag(flag)
            elements[name] = create_score_element_metadata(func)

    elements.freeze()
    return elements


def create_score_element_metadata(element: callable) -> Union[FunctionMetadata, EventLogMetadata]:
    flags = get_score_flag(element)

    if flags & ScoreFlag.EVENTLOG:
        return EventLogMetadata(element)
    else:
        return FunctionMetadata(element)


def get_score_flag(obj: callable, default: ScoreFlag = ScoreFlag.NONE) -> ScoreFlag:
    return getattr(obj, CONST_SCORE_FLAG, default)


def set_score_flag(obj: callable, flag: ScoreFlag) -> ScoreFlag:
    setattr(obj, CONST_SCORE_FLAG, flag)
    return flag


def set_score_flag_on(obj: callable, flag: ScoreFlag) -> ScoreFlag:
    flag |= get_score_flag(obj)
    set_score_flag(obj, flag)
    return flag


def is_all_score_flag_on(obj: callable, flag: ScoreFlag) -> bool:
    return utils.is_all_flag_on(get_score_flag(obj), flag)


def is_any_score_flag_on(obj: callable, flag: ScoreFlag) -> bool:
    return utils.is_any_flag_on(get_score_flag(obj), flag)


def get_score_element_metadata(score, func_name: str) -> ScoreElementMetadata:
    try:
        elements = getattr(score, CONST_CLASS_ELEMENT_METADATAS)
        return elements[func_name]
    except KeyError:
        raise MethodNotFoundException(
            f"Method not found: {type(score).__name__}.{func_name}")


def is_struct_valid(type_hint: type) -> bool:
    try:
        check_if_struct_is_valid(type_hint)
    except:
        return False

    return True


def check_if_struct_is_valid(type_hint: type, structs: Set[Any] = None):
    if structs is None:
        structs = set()

    if type_hint in structs:
        raise IllegalFormatException(f"Circular type hint: {type_hint}")

    structs.add(type_hint)
    annotations = get_annotations(type_hint, None)

    size = len(annotations)
    if size > _MAX_STRUCT_FIELDS:
        raise IllegalFormatException(
            f"Too many fields in struct: {size} > {_MAX_STRUCT_FIELDS}")

    for type_hint in annotations.values():
        origin = get_origin(type_hint)
        if is_base_type(origin):
            continue

        if origin is list:
            normalize_list_type_hint(type_hint)
        elif origin is dict:
            normalize_dict_type_hint(type_hint)
        elif origin is Union:
            normalize_union_type_hint(type_hint)
        elif is_struct(origin):
            check_if_struct_is_valid(type_hint, structs)
        else:
            raise IllegalFormatException(f"Invalid type hint: {type_hint}")


def check_score_flag(score_class: Union[type, object]):
    """Check for score flags strictly

    :param score_class:
    :return:
    """
    elements = getattr(score_class, CONST_CLASS_ELEMENT_METADATAS)
    if elements is None:
        raise IllegalFormatException(f"Invalid score class: {score_class}")

    for metadata in elements.values():
        verify_score_flag(metadata.flag, allow_payable_only=False)



# -*- coding: utf-8 -*-

# Copyright 2019 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING, Optional

from ..base.exception import InvalidParamsException
from ..icon_constant import Revision

if TYPE_CHECKING:
    from .coin_part import CoinPart
    from .delegation_part import DelegationPart
    from .stake_part import StakePart
    from ..base.address import Address
    from ..iconscore.icon_score_context import IconScoreContext


class Account(object):
    def __init__(self, address: 'Address', current_block_height: int, revision: int, *,
                 coin_part: Optional['CoinPart'] = None,
                 stake_part: Optional['StakePart'] = None,
                 delegation_part: Optional['DelegationPart'] = None):
        self._address: 'Address' = address
        self._current_block_height: int = current_block_height

        self._coin_part: 'CoinPart' = coin_part
        self._stake_part: 'StakePart' = stake_part
        self._delegation_part: 'DelegationPart' = delegation_part

        self.normalize(revision)

    @property
    def address(self):
        return self._address

    @property
    def coin_part(self) -> 'CoinPart':
        return self._coin_part

    @property
    def stake_part(self) -> 'StakePart':
        return self._stake_part

    @property
    def delegation_part(self) -> 'DelegationPart':
        return self._delegation_part

    @property
    def balance(self) -> int:
        if self.coin_part:
            return self.coin_part.balance
        raise InvalidParamsException("Invalid intend: coin_part is None")

    @property
    def stake(self) -> int:
        if self.stake_part:
            return self.stake_part.stake
        raise InvalidParamsException("Invalid intend: stake_part is None")

    @property
    def voting_weight(self) -> int:
        if self.stake_part:
            return self.stake_part.voting_weight
        raise InvalidParamsException("Invalid intend: stake_part is None")

    @property
    def unstake(self) -> int:
        if self.stake_part:
            return self.stake_part.unstake
        raise InvalidParamsException("Invalid intend: stake_part is None")

    @property
    def total_stake(self) -> int:
        if self.stake_part:
            return self.stake_part.total_stake
        raise InvalidParamsException("Invalid intend: stake_part is None")

    @property
    def unstake_block_height(self) -> int:
        if self.stake_part:
            return self.stake_part.unstake_block_height
        raise InvalidParamsException("Invalid intend: start_part is None")

    @property
    def unstakes_info(self) -> Optional[list]:
        if self.stake_part:
            return self.stake_part.unstakes_info

    @property
    def delegated_amount(self) -> int:
        if self.delegation_part:
            return self.delegation_part.delegated_amount
        raise InvalidParamsException("Invalid intend: delegation_part is None")

    @property
    def delegations(self) -> Optional[list]:
        if self.delegation_part:
            return self.delegation_part.delegations
        raise InvalidParamsException("Invalid intend: delegation_part is None")

    @property
    def delegations_amount(self) -> int:
        if self.delegation_part:
            return self.delegation_part.delegations_amount
        raise InvalidParamsException("Invalid intend: delegation_part is None")

    @property
    def voting_power(self) -> int:
        if self.stake_part and self.delegation_part:
            return self.stake_part.voting_weight - self.delegation_part.delegations_amount
        raise InvalidParamsException("Invalid intend: stake_part or delegation_part is None")

    def deposit(self, value: int):
        if self.coin_part is None:
            raise InvalidParamsException('Failed to delegation: InvalidAccount')

        self.coin_part.deposit(value)

    def withdraw(self, value: int):
        if self.coin_part is None:
            raise InvalidParamsException('Failed to delegation: InvalidAccount')

        self.coin_part.withdraw(value)

    def normalize(self, revision: int):
        if self.coin_part is None or self.stake_part is None:
            return

        balance: int = self.stake_part.normalize(self._current_block_height, revision)
        if balance > 0:
            if self.coin_part is None:
                raise InvalidParamsException('Failed to normalize: no coin part')

            self.coin_part.toggle_has_unstake(False)
            self.coin_part.deposit(balance)

    def set_stake(self, context: "IconScoreContext", value: int, unstake_lock_period: int):
        if self.coin_part is None or self.stake_part is None:
            raise InvalidParamsException('Failed to stake: InvalidAccount')

        if not isinstance(value, int) or value < 0:
            raise InvalidParamsException('Failed to stake: value is not int type or value < 0')

        total: int = self.balance + self.total_stake

        if total < value:
            raise InvalidParamsException(f'Failed to stake: total{total} < stake{value}')

        offset: int = value - self.total_stake

        if offset == 0:
            self.stake_part.reset_unstake()
        elif offset > 0:
            self.coin_part.withdraw(offset)
            self.stake_part.add_stake(offset)
            self.stake_part.reset_unstake()
        else:
            unlock_block_height: int = self._current_block_height + unstake_lock_period
            self.coin_part.toggle_has_unstake(True)
            if context.revision >= Revision.MULTIPLE_UNSTAKE.value:
                self.stake_part.set_unstakes_info(unlock_block_height,
                                                  -offset, context.unstake_slot_max)
            else:
                self.stake_part.set_unstake(unlock_block_height,  -offset)

    def update_delegated_amount(self, offset: int):
        if self.delegation_part is None:
            raise InvalidParamsException('Failed to delegation: InvalidAccount')

        self.delegation_part.update_delegated_amount(offset)

    def set_delegations(self, new_delegations: list):
        if self.delegation_part is None:
            raise InvalidParamsException('Failed to delegation: InvalidAccount')
        
        self.delegation_part.set_delegations(new_delegations)

    def __eq__(self, other) -> bool:
        """operator == overriding

        :param other: (CoinPart)
        """
        return isinstance(other, Account) \
            and self._address == other.address \
            and self._coin_part == other.coin_part \
            and self._stake_part == other.stake_part \
            and self._delegation_part == other.delegation_part

    def __ne__(self, other) -> bool:
        """operator != overriding

        :param other: (CoinPart)
        """
        return not self.__eq__(other)



# -*- coding: utf-8 -*-

# Copyright 2019 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABCMeta, abstractmethod
from typing import List

from ..iconscore.icon_score_context import IconScoreContext
from ..icx.icx_account import Account


class EngineListener(metaclass=ABCMeta):
    @abstractmethod
    def on_set_stake(self, context: 'IconScoreContext', account: 'Account'):
        pass

    @abstractmethod
    def on_set_delegation(self, context: 'IconScoreContext', delegated_accounts: List['Account']):
        pass



# -*- coding: utf-8 -*-

# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum, unique
from typing import TYPE_CHECKING, Optional

from ..base.address import Address

if TYPE_CHECKING:
    pass


@unique
class TraceType(Enum):
    CALL = 0
    REVERT = 1
    THROW = 2
    DESTROY = 3


class Trace(object):
    """
    A DataClass of a Trace.
    sample
    {
        "scoreAddress": "cx00...",
        "trace": "CALL",
        "data": ["cx11...", "transfer", ["hx00...", "0x1234"]]
    }

    """

    def __init__(
            self,
            score_address: 'Address',
            trace: TraceType,
            data: list = None) -> None:
        """
        Constructor

        :param score_address: an address of SCORE in which the trace is occurred
        :param trace: trace type
        :param data: a list of arguments or call function name
          e.g.
            transfer: [RECIPIENT, AMOUNT]
            call: [SCORE_ADDRESS_TO_CALL, METHOD, [ARGS_OF_METHOD]]
            revert: [CODE, MESSAGE]
            throw: [CODE, MESSAGE]
        """
        self.score_address: 'Address' = score_address
        self.trace: TraceType = trace
        self.data: list = data

    def __str__(self) -> str:
        return '\n'.join([f'{k}: {v}' for k, v in self.__dict__.items()])

    def to_dict(self, casing: Optional = None) -> dict:
        """
        Returns properties as `dict`
        :return: a dict
        """
        new_dict = {}
        for key, value in self.__dict__.items():
            if value is None:
                # Excludes properties which have `None` value
                continue

            if isinstance(value, TraceType):
                value = value.name

            new_dict[casing(key) if casing else key] = value

        return new_dict



# -*- coding: utf-8 -*-

# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
from abc import ABC
from collections import OrderedDict
from typing import TYPE_CHECKING, Optional, List, Iterable

from iconcommons.logger import Logger

from .icon_score_mapper import IconScoreMapper
from .icon_score_step import IconScoreStepCounter
from .icon_score_trace import Trace
from ..base.block import Block
from ..base.exception import FatalException, AccessDeniedException
from ..base.message import Message
from ..base.transaction import Transaction
from ..database.batch import BlockBatch, TransactionBatch
from ..icon_constant import (
    IconScoreContextType, IconScoreFuncType, TERM_PERIOD, PRepGrade, PREP_MAIN_PREPS, PREP_MAIN_AND_SUB_PREPS,
    TermFlag, PRepStatus,
    Revision, PRepFlag, RevisionChangedFlag, UNSTAKE_SLOT_MAX)
from ..icx.issue.regulator import Regulator

if TYPE_CHECKING:
    from .icon_score_base import IconScoreBase
    from .icon_score_event_log import EventLog

    from ..base.address import Address
    from ..prep.data import PRep, PRepContainer, Term
    from ..utils import ContextEngine, ContextStorage
    from ..prep.prep_address_converter import PRepAddressConverter
    from ..inv.container import Container as INVContainer
    from ..database.batch import Batch


class IconScoreContext(ABC):
    TAG = "CTX"

    score_root_path: str = None
    icon_score_mapper: 'IconScoreMapper' = None
    icon_service_flag: int = 0
    legacy_tbears_mode: bool = False
    iiss_initial_irep: int = 0

    engine: 'ContextEngine' = None
    storage: 'ContextStorage' = None

    main_prep_count: int = PREP_MAIN_PREPS
    main_and_sub_prep_count: int = PREP_MAIN_AND_SUB_PREPS
    term_period: int = TERM_PERIOD

    decentralize_trigger: float = 0

    precommitdata_log_flag = False
    step_trace_flag: bool = False
    log_level: str = None
    unstake_slot_max: int = UNSTAKE_SLOT_MAX

    """Contains the useful information to process user's JSON-RPC request
    """

    def __init__(self, context_type: 'IconScoreContextType' = IconScoreContextType.QUERY) -> None:
        """Constructor

        :param context_type: IconScoreContextType.GENESIS, INVOKE, QUERY
        """
        self.type: 'IconScoreContextType' = context_type
        # The type of external function which is called latest
        self.func_type: 'IconScoreFuncType' = IconScoreFuncType.WRITABLE
        self.block: Optional['Block'] = None
        self.tx: Optional['Transaction'] = None
        self.msg: Optional['Message'] = None
        self.current_address: Optional['Address'] = None
        self.block_batch: Optional['BlockBatch'] = None
        self.tx_batch: Optional['TransactionBatch'] = None
        # For 2-depth block invocation
        self._prev_block_batches: Optional[List['BlockBatch']] = \
            [] if context_type == IconScoreContextType.INVOKE else None
        self.rc_block_batch: list = []
        self.rc_tx_batch: list = []
        self.new_icon_score_mapper: Optional['IconScoreMapper'] = None
        self.cumulative_step_used: int = 0
        self.event_logs: Optional[List['EventLog']] = None
        self.traces: Optional[List['Trace']] = None
        self.fee_sharing_proportion = 0  # The proportion of fee by SCORE in percent (0-100)
        self.step_counter: Optional['IconScoreStepCounter'] = None

        self.msg_stack = []
        self.event_log_stack = []

        # PReps to update on invoke
        self._preps: Optional['PRepContainer'] = None
        self._tx_dirty_preps: Optional[OrderedDict['Address', 'PRep']] = None
        # Collect Main and Sub P-Reps which have just been invalidated by penalty or unregister
        # to use for updating term info at the end of invoke
        self._term: Optional['Term'] = None

        self._prep_address_converter: Optional['PRepAddressConverter'] = None
        self._inv_container: Optional['INVContainer'] = None
        self.regulator: Optional['Regulator'] = None
        self.revision_changed_flag: 'RevisionChangedFlag' = RevisionChangedFlag.NONE

    @classmethod
    def set_decentralize_trigger(cls, decentralize_trigger: float):
        decentralize_trigger: float = decentralize_trigger

        if not 1.0 > decentralize_trigger >= 0:
            raise FatalException(f"Invalid min delegation percent for decentralize: {decentralize_trigger}."
                                 f"Do not exceed 100% or negative value")
        IconScoreContext.decentralize_trigger = decentralize_trigger

    @property
    def revision(self) -> int:
        if self._inv_container:
            return self._inv_container.revision_code
        else:
            return 0

    @property
    def readonly(self):
        return self.type == IconScoreContextType.QUERY or self.func_type == IconScoreFuncType.READONLY

    @property
    def total_supply(self):
        return self.storage.icx.get_total_supply(self)

    @property
    def preps(self) -> Optional['PRepContainer']:
        return self._preps

    @property
    def term(self) -> Optional['Term']:
        return self._term

    @property
    def prep_address_converter(self) -> Optional['PRepAddressConverter']:
        return self._prep_address_converter

    @property
    def inv_container(self) -> Optional['INVContainer']:
        return self._inv_container

    def is_revision_changed(self, target_rev: int) -> bool:
        old: 'INVContainer' = self.engine.inv.inv_container
        new: 'INVContainer' = self.inv_container

        return old.revision_code != new.revision_code and new.revision_code == target_rev

    def get_batches(self) -> Iterable['Batch']:
        """Used to support 2-depth block invocation
        It is called in ContextDatabase.get_from_batch() on estimation or invoke

        Searching order: tx_batch -> block_batch -> prev_block_batch -> state_db
        """
        yield self.tx_batch
        yield self.block_batch

        # If contex.type is not INVOKE, self._prev_block_batches is None
        if self._prev_block_batches:
            for prev_block_batch in self._prev_block_batches:
                yield prev_block_batch

    def is_decentralized(self) -> bool:
        return self._term is not None

    def is_the_first_block_on_decentralization(self) -> bool:
        return self._term.sequence == 0 and self.block.height == self._term.start_block_height

    def set_func_type_by_icon_score(self, icon_score: 'IconScoreBase', func_name: str):
        is_func_readonly = getattr(icon_score, '_IconScoreBase__is_func_readonly')
        if func_name is not None and is_func_readonly(func_name):
            self.func_type = IconScoreFuncType.READONLY
        else:
            self.func_type = IconScoreFuncType.WRITABLE

    # TODO should remove after update GOVERNANCE 0.0.6 afterward
    def deploy(self, tx_hash: bytes) -> None:
        warnings.warn("legacy function don't use.", DeprecationWarning, stacklevel=2)
        self.engine.deploy.deploy(self, tx_hash)

    def update_batch(self):
        """Called when a transaction is done

        :return:
        """

        # Call update_dirty_prep_batch before update_state_db_batch()
        self.update_dirty_prep_batch()
        self.update_state_db_batch()
        self.update_rc_db_batch()

    def update_state_db_batch(self):
        self.block_batch.update(self.tx_batch)
        self.tx_batch.clear()

    def update_rc_db_batch(self):
        self.rc_block_batch.extend(self.rc_tx_batch)
        self.rc_tx_batch.clear()

    def update_dirty_prep_batch(self):
        """Apply updated P-Rep data to context.preps every time when a tx is done

        Caution: call update_dirty_prep_batch before update_state_db_batch()
        """
        for dirty_prep in self._tx_dirty_preps.values():
            if self._term is not None:
                self._update_term(dirty_prep)

            self._update_prep_address_converter(dirty_prep=dirty_prep)

            self._preps.replace(dirty_prep)
            # Write serialized dirty_prep data into tx_batch
            self.storage.prep.put_prep(self, dirty_prep)
            dirty_prep.freeze()

        self._tx_dirty_preps.clear()

    def _update_prep_address_converter(self, dirty_prep: 'PRep'):
        if not self._preps.contains(dirty_prep.address, active_prep_only=False):
            # registerPRep
            self._prep_address_converter.add_node_address(node=dirty_prep.node_address,
                                                          prep=dirty_prep.address)
        elif dirty_prep.is_flags_on(PRepFlag.NODE_ADDRESS):
            # setPRep
            old_prep = self._preps.get_by_address(dirty_prep.address)
            self._prep_address_converter.replace_node_address(node=dirty_prep.node_address,
                                                              prep=dirty_prep.address,
                                                              prev_node=old_prep.node_address)
        elif dirty_prep.status != PRepStatus.ACTIVE:
            # unregisterPRep or disqualified by productivity penalty
            self._prep_address_converter.delete_node_address(node=dirty_prep.node_address,
                                                             prep=dirty_prep.address)

    def _update_term(self, dirty_prep: 'PRep'):
        """Update term info with dirty_prep

        :param dirty_prep:
        :return:
        """
        # If dirty_prep address is not contained in term, no need to update self._term
        if dirty_prep.address not in self._term:
            return

        if dirty_prep.is_electable():
            self._update_term_flag(dirty_prep)
        else:
            self._remove_invalid_elected_prep_from_term(dirty_prep)

    def _update_term_flag(self, dirty_prep: 'PRep'):
        """Notify the term that p2p_endpoint of a main P-Rep is changed

        :param dirty_prep: dirty prep
        """
        if self.revision < Revision.REALTIME_P2P_ENDPOINT_UPDATE.value:
            return

        if not self._term.is_main_prep(dirty_prep.address):
            return

        if dirty_prep.flags & (PRepFlag.P2P_ENDPOINT | PRepFlag.NODE_ADDRESS):
            self._duplicate_term()
            if dirty_prep.is_flags_on(PRepFlag.P2P_ENDPOINT):
                self._term.on_main_prep_changed(TermFlag.MAIN_PREP_P2P_ENDPOINT)
            if dirty_prep.is_flags_on(PRepFlag.NODE_ADDRESS):
                self._term.on_main_prep_changed(TermFlag.MAIN_PREP_NODE_ADDRESS)
            Logger.info(tag=self.TAG, msg=f"_update_term_flag: {dirty_prep}")
        else:
            Logger.info(tag=self.TAG, msg=f"_update_term_flag(x): {dirty_prep}")

    def _remove_invalid_elected_prep_from_term(self, dirty_prep: 'PRep'):
        """Remove an invalidated elected P-Rep from the current term

        :param dirty_prep: dirty prep
        """
        self._duplicate_term()

        # Just in case, reset the P-Rep grade one to CANDIDATE
        dirty_prep.grade = PRepGrade.CANDIDATE

        self._term.update_invalid_elected_preps([dirty_prep])

        Logger.info(tag=self.TAG, msg=f"Invalid main and sub prep: {dirty_prep}")

    def _duplicate_term(self) -> Optional['Term']:
        if self.type not in (IconScoreContextType.INVOKE, IconScoreContextType.ESTIMATION):
            raise AccessDeniedException(f"Method not allowed: context={self.type.name}")

        if self._term is not None and self._term.is_frozen():
            self._term = self._term.copy()

        return self._term

    def clear_batch(self):
        if self.tx_batch:
            self.tx_batch.clear()
        if self.rc_tx_batch:
            self.rc_tx_batch.clear()
        if self._tx_dirty_preps:
            self._tx_dirty_preps.clear()

    def get_prep(self, address: 'Address', mutable: bool = False) -> Optional['PRep']:
        prep: Optional['PRep'] = None

        if self._tx_dirty_preps is not None:
            prep = self._tx_dirty_preps.get(address)

        if prep is None:
            prep = self._preps.get_by_address(address)

        if prep and prep.is_frozen() and mutable:
            prep = prep.copy()

        return prep

    def put_dirty_prep(self, prep: 'PRep'):
        # Logger.debug(tag=self.TAG, msg=f"put_dirty_prep() start: {prep}")

        if self._tx_dirty_preps is None:
            Logger.warning(tag=self.TAG, msg="self._tx_dirty_preps is None")
            # Logger.debug(tag=self.TAG, msg="put_dirty_prep() end")
            return

        if not prep.is_dirty() and self.revision >= Revision.OPTIMIZE_DIRTY_PREP_UPDATE.value:
            Logger.info(tag=self.TAG, msg=f"No need to update an unchanged P-Rep: revision={self.revision}")
            # Logger.debug(tag=self.TAG, msg="put_dirty_prep() end")
            return

        self._tx_dirty_preps[prep.address] = prep

        # Logger.debug(tag=self.TAG, msg="put_dirty_prep() end")

    def set_step_counter(self, step_limit: Optional[int] = None):
        if self.type == IconScoreContextType.ESTIMATION:
            context_type: 'IconScoreContextType' = IconScoreContextType.INVOKE
        else:
            context_type: 'IconScoreContextType' = self.type

        max_step_limit: int = self._inv_container.max_step_limits.get(context_type)
        if step_limit:
            step_limit: int = min(step_limit, max_step_limit)
        else:
            step_limit: int = max_step_limit

        self.step_counter = IconScoreStepCounter(step_price=self._inv_container.step_price,
                                                 step_costs=self._inv_container.step_costs,
                                                 step_limit=step_limit,
                                                 step_trace_flag=self._is_step_trace_on())

    def _is_step_trace_on(self) -> bool:
        return self.step_trace_flag and self.type == IconScoreContextType.INVOKE


class IconScoreContextFactory(object):
    def __init__(self):
        pass

    def create(self,
               context_type: 'IconScoreContextType',
               block: 'Block',
               prev_block_batches: Iterable['Batch'] = None):
        context: 'IconScoreContext' = self._create_context(context_type)
        context.block = block

        if context_type == IconScoreContextType.DIRECT:
            return context

        # For 2-depth block invocation
        if prev_block_batches:
            context._prev_block_batches = [batch for batch in prev_block_batches]
        self._set_context_attributes_for_processing_tx(context)
        return context

    @classmethod
    def _create_context(cls, context_type: 'IconScoreContextType') -> 'IconScoreContext':
        return IconScoreContext(context_type)

    @classmethod
    def _set_context_attributes_for_processing_tx(cls, context: 'IconScoreContext'):
        if context.type in (IconScoreContextType.INVOKE, IconScoreContextType.ESTIMATION):
            context.block_batch = BlockBatch(Block.from_block(context.block))
            context.tx_batch = TransactionBatch()

            context.new_icon_score_mapper = IconScoreMapper()

            # For PRep management
            context._preps = context.engine.prep.preps.copy(mutable=True)
            context._tx_dirty_preps = OrderedDict()
            container: 'INVContainer' = context.engine.inv.inv_container.copy()
            context._inv_container = container
            context._prep_address_converter = context.engine.prep.prep_address_converter.copy()
        else:
            # Readonly
            context._preps = context.engine.prep.preps
            context._inv_container = context.engine.inv.inv_container
            context._prep_address_converter = context.engine.prep.prep_address_converter

        context._term = context.engine.prep.term



# -*- coding: utf-8 -*-

# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING

from ..base.address import Address
from ..base.exception import InvalidParamsException
from ..icon_constant import Revision
from ..utils import is_builtin_score

if TYPE_CHECKING:
    from .icon_score_base import IconScoreBase
    from ..database.db import IconScoreDatabase


class IconScoreInfo(object):
    """Contains information on one icon score

    If this class is not necessary anymore, Remove it
    """

    def __init__(self, score_class: type, score_db: 'IconScoreDatabase', tx_hash: bytes) -> None:
        """Constructor

        :param score_class:
        :param score_db:
        :param tx_hash:
        """
        self._tx_hash = tx_hash
        self._score_class = score_class
        self._score_db = score_db
        self._score = None

    @property
    def tx_hash(self) -> bytes:
        return self._tx_hash

    @property
    def score_class(self) -> type:
        return self._score_class

    @property
    def score_db(self) -> 'IconScoreDatabase':
        return self._score_db

    @property
    def address(self) -> 'Address':
        return self._score_db.address

    def get_score(self, revision: int) -> 'IconScoreBase':
        """Provide a score instance according to the revision.
        1. revision <= 2: Returns a cached score instance
        2. revision > 2: Returns a newly created score instance

        :param revision:
        :return:
        """
        if revision <= Revision.TWO.value or is_builtin_score(str(self.address)):
            if self._score is None:
                self._score = self.create_score()

            return self._score

        return self.create_score()

    def create_score(self) -> 'IconScoreBase':
        return self._score_class(self._score_db)


class IconScoreMapperObject(dict):
    def __getitem__(self, key: 'Address') -> 'IconScoreInfo':
        """operator[] overriding

        :param key:
        :return: IconScoreInfo instance
        """
        self._check_key_type(key)
        return super().__getitem__(key)

    def __setitem__(self,
                    key: 'Address',
                    value: 'IconScoreInfo') -> None:
        """
        :param key:
        :param value: IconScoreInfo
        """
        self._check_key_type(key)
        self._check_value_type(value)
        super().__setitem__(key, value)

    @staticmethod
    def _check_key_type(address: 'Address') -> None:
        """Check if key type is an icon score address type or not.

        :param address: icon score address
        """
        if not isinstance(address, Address):
            raise InvalidParamsException(
                f'{address} is an invalid address')
        if not address.is_contract:
            raise InvalidParamsException(
                f'{address} is not an icon score address.')

    @staticmethod
    def _check_value_type(info: 'IconScoreInfo') -> None:
        if not isinstance(info, IconScoreInfo):
            raise InvalidParamsException(
                f'{info} is not IconScoreInfo type.')



# -*- coding: utf-8 -*-

# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from iconservice import *

TAG = 'Governance'
DEBUG = False

CURRENT = 'current'
NEXT = 'next'
STATUS = 'status'
DEPLOY_TX_HASH = 'deployTxHash'
AUDIT_TX_HASH = 'auditTxHash'
VALID_STATUS_KEYS = [STATUS, DEPLOY_TX_HASH, AUDIT_TX_HASH]

STATUS_PENDING = 'pending'
STATUS_ACTIVE = 'active'
STATUS_INACTIVE = 'inactive'
STATUS_REJECTED = 'rejected'

STEP_TYPE_DEFAULT = 'default'
STEP_TYPE_CONTRACT_CALL = 'contractCall'
STEP_TYPE_CONTRACT_CREATE = 'contractCreate'
STEP_TYPE_CONTRACT_UPDATE = 'contractUpdate'
STEP_TYPE_CONTRACT_DESTRUCT = 'contractDestruct'
STEP_TYPE_CONTRACT_SET = 'contractSet'
STEP_TYPE_GET = 'get'
STEP_TYPE_SET = 'set'
STEP_TYPE_REPLACE = 'replace'
STEP_TYPE_DELETE = 'delete'
STEP_TYPE_INPUT = 'input'
STEP_TYPE_EVENT_LOG = 'eventLog'
STEP_TYPE_API_CALL = 'apiCall'
INITIAL_STEP_COST_KEYS = [STEP_TYPE_DEFAULT,
                          STEP_TYPE_CONTRACT_CALL, STEP_TYPE_CONTRACT_CREATE, STEP_TYPE_CONTRACT_UPDATE,
                          STEP_TYPE_CONTRACT_DESTRUCT, STEP_TYPE_CONTRACT_SET,
                          STEP_TYPE_GET, STEP_TYPE_SET, STEP_TYPE_REPLACE, STEP_TYPE_DELETE, STEP_TYPE_INPUT,
                          STEP_TYPE_EVENT_LOG, STEP_TYPE_API_CALL]

CONTEXT_TYPE_INVOKE = 'invoke'
CONTEXT_TYPE_QUERY = 'query'


class StepCosts:
    """
    DB for stepCosts management.
    It is combined DictDB and ArrayDB in order to iterate items.
    """
    _STEP_TYPES = 'step_types'
    _STEP_COSTS = 'step_costs'

    def __init__(self, db: 'IconScoreDatabase'):
        self._step_types = ArrayDB(self._STEP_TYPES, db, value_type=str)
        self._step_costs = DictDB(self._STEP_COSTS, db, value_type=int)

    def __setitem__(self, step_type: str, cost: int):
        if step_type not in self._step_costs:
            self._step_types.put(step_type)

        self._step_costs[step_type] = cost

    def __getitem__(self, step_type: str):
        return self._step_costs[step_type]

    def __delitem__(self, step_type: str):
        # delete does not actually do delete but set zero
        if step_type in self._step_costs:
            self._step_costs[step_type] = 0

    def __contains__(self, step_type: str):
        return step_type in self._step_costs

    def __iter__(self):
        return self._step_types.__iter__()

    def __len__(self):
        return self._step_types.__len__()

    def items(self):
        for step_type in self._step_types:
            yield (step_type, self._step_costs[step_type])


class Governance(IconScoreBase):

    _SCORE_STATUS = 'score_status'
    _AUDITOR_LIST = 'auditor_list'
    _DEPLOYER_LIST = 'deployer_list'
    _SCORE_BLACK_LIST = 'score_black_list'
    _STEP_PRICE = 'step_price'
    _MAX_STEP_LIMITS = 'max_step_limits'

    @eventlog(indexed=1)
    def Accepted(self, tx_hash: str):
        pass

    @eventlog(indexed=1)
    def Rejected(self, tx_hash: str, reason: str):
        pass

    @eventlog(indexed=1)
    def StepPriceChanged(self, step_price: int):
        pass

    @eventlog(indexed=1)
    def StepCostChanged(self, step_type: str, cost: int):
        pass

    def __init__(self, db: IconScoreDatabase) -> None:
        super().__init__(db)
        self._score_status = DictDB(self._SCORE_STATUS, db, value_type=bytes, depth=3)
        self._auditor_list = ArrayDB(self._AUDITOR_LIST, db, value_type=Address)
        self._deployer_list = ArrayDB(self._DEPLOYER_LIST, db, value_type=Address)
        self._score_black_list = ArrayDB(self._SCORE_BLACK_LIST, db, value_type=Address)
        self._step_price = VarDB(self._STEP_PRICE, db, value_type=int)
        self._step_costs = StepCosts(db)
        self._max_step_limits = DictDB(self._MAX_STEP_LIMITS, db, value_type=int)

    def on_install(self,
                   stepPrice: int = 10 ** 10,
                   maxInvokeStepLimit: int = 0x78000000,
                   maxQueryStepLimit: int = 0x780000) -> None:
        super().on_install()
        # add owner into initial auditor list
        Logger.debug(f'on_install: owner = "{self.owner}"', TAG)
        self._auditor_list.put(self.owner)
        # add owner into initial deployer list
        self._deployer_list.put(self.owner)
        # set initial step price
        self._step_price.set(stepPrice)
        # set initial step costs
        self._set_initial_step_costs()
        # set initial max step limits
        self._max_step_limits[CONTEXT_TYPE_INVOKE] = maxInvokeStepLimit
        self._max_step_limits[CONTEXT_TYPE_QUERY] = maxQueryStepLimit

    def on_update(self) -> None:
        super().on_update()

        if len(self._step_costs) == 0:
            # migrates from old DB of step_costs.
            for step_type in INITIAL_STEP_COST_KEYS:
                if step_type in self._step_costs:
                    self._step_costs._step_types.put(step_type)

    def _get_current_status(self, score_address: Address):
        return self._score_status[score_address][CURRENT]

    def _get_next_status(self, score_address: Address):
        return self._score_status[score_address][NEXT]

    @staticmethod
    def _fill_status_with_str(db: DictDB):
        count = 0
        status = {}
        for key in VALID_STATUS_KEYS:
            value: bytes = db[key]
            if value:
                if key == STATUS:
                    status[key] = value.decode()
                else:
                    status[key] = value
                count += 1
        return count, status

    @staticmethod
    def _save_status(db: DictDB, status: dict) -> None:
        for key in VALID_STATUS_KEYS:
            value: bytes = status[key]
            if value:
                db[key] = value

    @staticmethod
    def _remove_status(db: DictDB) -> None:
        for key in VALID_STATUS_KEYS:
            value = db[key]
            if value:
                del db[key]

    @external(readonly=True)
    def getScoreStatus(self, address: Address) -> dict:
        # check score address
        current_tx_hash, next_tx_hash = self.get_tx_hashes_by_score_address(address)
        if current_tx_hash is None and next_tx_hash is None:
            self.revert('SCORE not found')
        result = {}
        build_initial_status = False
        # get current status
        _current = self._get_current_status(address)
        count1, status = self._fill_status_with_str(_current)
        if count1 > 0:
            if current_tx_hash is None:
                self.revert('current_tx_hash is None')
            if current_tx_hash != status[DEPLOY_TX_HASH]:
                self.revert('Current deploy tx mismatch')
            # audit has been performed (accepted)
            result[CURRENT] = status
        # get next status
        _next = self._get_next_status(address)
        count2, status = self._fill_status_with_str(_next)
        if count2 > 0:
            # check if another pending tx has been arrived
            if next_tx_hash is not None and \
                    next_tx_hash != status[DEPLOY_TX_HASH]:
                build_initial_status = True
            else:
                # audit has been performed (rejected)
                result[NEXT] = status
        else:
            if next_tx_hash is not None:
                build_initial_status = True
        # there is no information, build initial status
        if build_initial_status:
            status = {
                STATUS: STATUS_PENDING,
                DEPLOY_TX_HASH: next_tx_hash
            }
            result[NEXT] = status
        return result

    @external(readonly=True)
    def getStepPrice(self) -> int:
        return self._step_price.get()

    @external
    def setStepPrice(self, stepPrice: int):
        # only owner can set new step price
        if self.msg.sender != self.owner:
            self.revert('Invalid sender: not owner')
        if stepPrice > 0:
            self._step_price.set(stepPrice)
            self.StepPriceChanged(stepPrice)

    @external
    def acceptScore(self, txHash: bytes):
        # check message sender
        Logger.debug(f'acceptScore: msg.sender = "{self.msg.sender}"', TAG)
        if self.msg.sender not in self._auditor_list:
            self.revert('Invalid sender: no permission')
        # check txHash
        score_address = self.get_score_address_by_tx_hash(txHash)
        if score_address is None:
            self.revert('Invalid txHash')
        Logger.debug(f'acceptScore: score_address = "{score_address}"', TAG)
        # check next: it should be 'pending'
        result = self.getScoreStatus(score_address)
        try:
            next_status = result[NEXT][STATUS]
            if next_status != STATUS_PENDING:
                self.revert(f'Invalid status: next is {next_status}')
        except KeyError:
            self.revert('Invalid status: no next status')
        # next: pending -> null
        _next = self._get_next_status(score_address)
        self._remove_status(_next)
        # current: null -> active
        _current = self._get_current_status(score_address)
        status = {
            STATUS: STATUS_ACTIVE,
            DEPLOY_TX_HASH: txHash,
            AUDIT_TX_HASH: self.tx.hash
        }
        self._save_status(_current, status)
        self.deploy(txHash)
        self.Accepted('0x' + txHash.hex())

    @external
    def rejectScore(self, txHash: bytes, reason: str):
        # check message sender
        Logger.debug(f'rejectScore: msg.sender = "{self.msg.sender}"', TAG)
        if self.msg.sender not in self._auditor_list:
            self.revert('Invalid sender: no permission')
        # check txHash
        score_address = self.get_score_address_by_tx_hash(txHash)
        if score_address is None:
            self.revert('Invalid txHash')
        Logger.debug(f'rejectScore: score_address = "{score_address}", reason = {reason}', TAG)
        # check next: it should be 'pending'
        result = self.getScoreStatus(score_address)
        try:
            next_status = result[NEXT][STATUS]
            if next_status != STATUS_PENDING:
                self.revert(f'Invalid status: next is {next_status}')
        except KeyError:
            self.revert('Invalid status: no next status')
        # next: pending -> rejected
        _next = self._get_next_status(score_address)
        status = {
            STATUS: STATUS_REJECTED,
            DEPLOY_TX_HASH: txHash,
            AUDIT_TX_HASH: self.tx.hash
        }
        self._save_status(_next, status)
        self.Rejected('0x' + txHash.hex(), reason)

    @external
    def addAuditor(self, address: Address):
        # check message sender, only owner can add new auditor
        if self.msg.sender != self.owner:
            self.revert('Invalid sender: not owner')
        if address not in self._auditor_list:
            self._auditor_list.put(address)
        if DEBUG is True:
            self._print_auditor_list('addAuditor')

    @external
    def removeAuditor(self, address: Address):
        if address not in self._auditor_list:
            self.revert('Invalid address: not in list')
        # check message sender
        if self.msg.sender != self.owner:
            if self.msg.sender != address:
                self.revert('Invalid sender: not yourself')
        # get the topmost value
        top = self._auditor_list.pop()
        if top != address:
            for i in range(len(self._auditor_list)):
                if self._auditor_list[i] == address:
                    self._auditor_list[i] = top
        if DEBUG is True:
            self._print_auditor_list('removeAuditor')

    def _print_auditor_list(self, header: str):
        Logger.debug(f'{header}: list len = {len(self._auditor_list)}', TAG)
        for auditor in self._auditor_list:
            Logger.debug(f' --- {auditor}', TAG)

    @external
    def addDeployer(self, address: Address):
        # check message sender, only owner can add new deployer
        if self.msg.sender != self.owner:
            self.revert('Invalid sender: not owner')
        if address not in self._deployer_list:
            self._deployer_list.put(address)
        if DEBUG is True:
            self._print_deployer_list('addDeployer')

    @external
    def removeDeployer(self, address: Address):
        if address not in self._deployer_list:
            self.revert('Invalid address: not in list')
        # check message sender
        if self.msg.sender != self.owner:
            if self.msg.sender != address:
                self.revert('Invalid sender: not yourself')
        # get the topmost value
        top = self._deployer_list.pop()
        if top != address:
            for i in range(len(self._deployer_list)):
                if self._deployer_list[i] == address:
                    self._deployer_list[i] = top
        if DEBUG is True:
            self._print_deployer_list('removeDeployer')

    @external(readonly=True)
    def isDeployer(self, address: Address) -> bool:
        Logger.debug(f'isDeployer address: {address}', TAG)
        return address in self._deployer_list

    def _print_deployer_list(self, header: str):
        Logger.debug(f'{header}: list len = {len(self._deployer_list)}', TAG)
        for deployer in self._deployer_list:
            Logger.debug(f' --- {deployer}', TAG)

    @external
    def addToScoreBlackList(self, address: Address):
        if not address.is_contract:
            self.revert(f'Invalid SCORE Address: {address}')

        # check message sender, only owner can add new blacklist
        if self.msg.sender != self.owner:
            self.revert('Invalid sender: not owner')
        if address not in self._score_black_list:
            self._score_black_list.put(address)
        if DEBUG is True:
            self._print_black_list('addScoreToBlackList')

    @external
    def removeFromScoreBlackList(self, address: Address):
        if address not in self._score_black_list:
            self.revert('Invalid address: not in list')

        # check message sender, only owner can remove from blacklist
        if self.msg.sender != self.owner:
            self.revert('Invalid sender: not owner')
        # get the topmost value
        top = self._score_black_list.pop()
        if top != address:
            for i in range(len(self._score_black_list)):
                if self._score_black_list[i] == address:
                    self._score_black_list[i] = top
        if DEBUG is True:
            self._print_black_list('removeScoreFromBlackList')

    @external(readonly=True)
    def isInScoreBlackList(self, address: Address) -> bool:
        Logger.debug(f'isInBlackList address: {address}', TAG)
        return address in self._score_black_list

    def _print_black_list(self, header: str):
        Logger.debug(f'{header}: list len = {len(self._score_black_list)}', TAG)
        for addr in self._score_black_list:
            Logger.debug(f' --- {addr}', TAG)

    def _set_initial_step_costs(self):
        initial_costs = {
            STEP_TYPE_DEFAULT: 1_000_000,
            STEP_TYPE_CONTRACT_CALL: 15_000,
            STEP_TYPE_CONTRACT_CREATE: 200_000,
            STEP_TYPE_CONTRACT_UPDATE: 80_000,
            STEP_TYPE_CONTRACT_DESTRUCT: -70_000,
            STEP_TYPE_CONTRACT_SET: 30_000,
            STEP_TYPE_GET: 0,
            STEP_TYPE_SET: 200,
            STEP_TYPE_REPLACE: 50,
            STEP_TYPE_DELETE: -150,
            STEP_TYPE_INPUT: 200,
            STEP_TYPE_EVENT_LOG: 100,
            STEP_TYPE_API_CALL: 0
        }
        for key, value in initial_costs.items():
            self._step_costs[key] = value

    @external(readonly=True)
    def getStepCosts(self) -> dict:
        result = {}
        for key, value in self._step_costs.items():
            result[key] = value
        return result

    @external
    def setStepCost(self, stepType: str, cost: int):
        # only owner can set new step cost
        if self.msg.sender != self.owner:
            self.revert('Invalid sender: not owner')
        if cost < 0:
            if stepType != STEP_TYPE_CONTRACT_DESTRUCT and \
                    stepType != STEP_TYPE_DELETE:
                self.revert(f'Invalid step cost: {stepType}, {cost}')
        self._step_costs[stepType] = cost
        self.StepCostChanged(stepType, cost)

    @external(readonly=True)
    def getMaxStepLimit(self, context_type: str) -> int:
        return self._max_step_limits[context_type]



# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import IntFlag, unique, IntEnum, Enum, auto, Flag

SYSTEM_ADDRESS = "cx0000000000000000000000000000000000000000"
GOVERNANCE_ADDRESS = "cx0000000000000000000000000000000000000001"

ICON_DEPLOY_LOG_TAG = "DEPLOY"
ICON_LOADER_LOG_TAG = "LOADER"
ICX_LOG_TAG = "ICX"
ICON_DB_LOG_TAG = "DB"
IISS_LOG_TAG = "IISS"
STEP_LOG_TAG = "STEP"
WAL_LOG_TAG = "WAL"
ROLLBACK_LOG_TAG = "ROLLBACK"
BACKUP_LOG_TAG = "BACKUP"

JSONRPC_VERSION = '2.0'
CHARSET_ENCODING = 'utf-8'

ICX_IN_LOOP = 10 ** 18
# 32bytes == 256bit
DEFAULT_BYTE_SIZE = 32
DATA_BYTE_ORDER = 'big'  # big endian
# Fixed fee is 0.01 icx.
FIXED_FEE = 10 ** 16
# Max data field size
MAX_DATA_SIZE = 512 * 1024

# Max external call count(1 is default SCORE call, 1024 is external call in the SCORE)
MAX_EXTERNAL_CALL_COUNT = 1 + 1024

# Max call stack size
MAX_CALL_STACK_SIZE = 64

ICON_DEX_DB_NAME = 'icon_dex'
PACKAGE_JSON_FILE = 'package.json'

ICX_TRANSFER_EVENT_LOG = 'ICXTransfer(Address,Address,int)'

ICON_SCORE_QUEUE_NAME_FORMAT = "IconScore.{channel_name}.{amqp_key}"
ICON_SERVICE_PROCTITLE_FORMAT = "icon_service." \
                                "{scoreRootPath}." \
                                "{stateDbRootPath}." \
                                "{channel}.{amqpKey}." \
                                "{amqpTarget}"

BUILTIN_SCORE_ADDRESS_MAPPER = {
    'governance': GOVERNANCE_ADDRESS,
    'system': SYSTEM_ADDRESS
}
BUILTIN_SCORE_IMPORT_WHITE_LIST = {"iconservice.iconscore.system": "['*']"}

ZERO_TX_HASH = bytes(DEFAULT_BYTE_SIZE)


class IssueDataKey:
    PREP = "prep"
    IREP = "irep"
    RREP = "rrep"

    EEP = "eep"
    IEEP = "ieep"
    REEP = "reep"

    DAPP = "dapp"
    IDAPP = "idapp"
    RDAPP = "rdapp"

    ISSUE_RESULT = "result"
    COVERED_BY_FEE = "coveredByFee"
    COVERED_BY_OVER_ISSUED_ICX = "coveredByOverIssuedICX"
    ISSUE = "issue"

    TOTAL_DELEGATION = "totalDelegation"
    VALUE = "value"

    TOTAL = "total"


ISSUE_EVENT_LOG_MAPPER = {
    IssueDataKey.PREP: {
        "event_signature": "PRepIssued(int,int,int,int)",
        "data": [IssueDataKey.IREP, IssueDataKey.RREP, IssueDataKey.TOTAL_DELEGATION,
                 IssueDataKey.VALUE]
    },
    IssueDataKey.TOTAL: {
        "event_signature": "ICXIssued(int,int,int,int)",
        "data": []
    }
}

ISSUE_CALCULATE_ORDER = [IssueDataKey.PREP]

BASE_TRANSACTION_INDEX = 0


class Revision(Enum):
    GENESIS = 0
    TWO = 2
    THREE = 3
    FOUR = 4
    IISS = 5
    DECENTRALIZATION = 6
    FIX_TOTAL_ELECTED_PREP_DELEGATED = 7

    # Revision 8
    REALTIME_P2P_ENDPOINT_UPDATE = 8
    OPTIMIZE_DIRTY_PREP_UPDATE = 8

    # Revision 9
    FIX_EMAIL_VALIDATION = 9
    DIVIDE_NODE_ADDRESS = 9
    FIX_BURN_EVENT_SIGNATURE = 9
    ADD_LOGS_BLOOM_ON_BASE_TX = 9
    SCORE_FUNC_PARAMS_CHECK = 9
    SYSTEM_SCORE_ENABLED = 9
    CHANGE_MAX_DELEGATIONS_TO_100 = 9
    PREVENT_DUPLICATED_ENDPOINT = 9
    SET_IREP_VIA_NETWORK_PROPOSAL = 9
    MULTIPLE_UNSTAKE = 9
    FIX_COIN_PART_BYTES_ENCODING = 9
    STRICT_SCORE_DECORATOR_CHECK = 9

    LATEST = 9


RC_DB_VERSION_0 = 0
RC_DB_VERSION_2 = 2

# The case that version is updated but not revision, set the version to the current revision
# The case that both version and revision is updated, add revision field to the version table
# The case that only revision is changed, do not update this table
RC_DATA_VERSION_TABLE = {
    Revision.IISS.value: RC_DB_VERSION_0,
    Revision.DECENTRALIZATION.value: RC_DB_VERSION_2
}

IISS_DB = 'iiss'
RC_SOCKET = 'iiss.sock'

META_DB = 'meta'


class ConfigKey:
    BUILTIN_SCORE_OWNER = 'builtinScoreOwner'
    SERVICE = 'service'
    SERVICE_FEE = 'fee'
    SERVICE_AUDIT = 'audit'
    SERVICE_SCORE_PACKAGE_VALIDATOR = 'scorePackageValidator'
    SCORE_ROOT_PATH = 'scoreRootPath'
    STATE_DB_ROOT_PATH = 'stateDbRootPath'
    CHANNEL = 'channel'
    AMQP_KEY = 'amqpKey'
    AMQP_TARGET = 'amqpTarget'
    CONFIG = 'config'
    TBEARS_MODE = 'tbearsMode'
    IISS_CALCULATE_PERIOD = "iissCalculatePeriod"
    TERM_PERIOD = 'termPeriod'
    INITIAL_IREP = 'initialIRep'
    PREP_MAIN_PREPS = 'mainPRepCount'
    PREP_MAIN_AND_SUB_PREPS = 'mainAndSubPRepCount'
    IPC_TIMEOUT = 'ipcTimeout'

    # log
    LOG = 'log'
    LOGGER = "logger"
    LOG_FILE_PATH = 'filePath'
    LOG_LEVEL = "level"
    LOG_OUTPUT_TYPE = "outputType"
    LOG_ROTATE = "rotate"
    LOG_ROTATE_TYPE = "type"
    LOG_ROTATE_PERIOD = "period"
    LOG_ROTATE_INTERVAL = "interval"
    LOG_ROTATE_AT_TIME = "atTime"
    LOG_ROTATE_MAX_BYTES = "maxBytes"
    LOG_ROTATE_BACKUP_COUNT = "backupCount"
    STEP_TRACE_FLAG = 'stepTraceFlag'
    PRECOMMIT_DATA_LOG_FLAG = 'precommitDataLogFlag'

    # Reward calculator
    # executable path
    ICON_RC_DIR_PATH = 'iconRcPath'
    # Boolean which determines Opening RC monitor channel (Default True)
    ICON_RC_MONITOR = 'iconRcMonitor'

    # IISS meta data
    IISS_META_DATA = "iissMetaData"
    REWARD_POINT = 'rewardPoint'
    REWARD_MIN = "rewardMin"
    REWARD_MAX = "rewardMAX"
    UN_STAKE_LOCK_MIN = "lockMin"
    UN_STAKE_LOCK_MAX = "lockMax"

    PREP_REGISTRATION_FEE = "prepRegistrationFee"

    DECENTRALIZE_TRIGGER = "decentralizeTrigger"
    PENALTY_GRACE_PERIOD = "penaltyGracePeriod"
    LOW_PRODUCTIVITY_PENALTY_THRESHOLD = "lowProductivityPenaltyThreshold"
    BLOCK_VALIDATION_PENALTY_THRESHOLD = "blockValidationPenaltyThreshold"

    # The maximum number of backup files for rollback
    BACKUP_FILES = "backupFiles"

    # Block invoke timeout in second
    BLOCK_INVOKE_TIMEOUT = "blockInvokeTimeout"

    UNSTAKE_SLOT_MAX = "unstakeSlotMax"


class EnableThreadFlag(IntFlag):
    INVOKE = 1
    QUERY = 2
    VALIDATE = 4


class IconServiceFlag(IntFlag):
    FEE = 1
    AUDIT = 2
    # DEPLOYER_WHITE_LIST = 4
    SCORE_PACKAGE_VALIDATOR = 8


class IconNetworkValueType(Enum):
    SERVICE_CONFIG = b'service_config'

    STEP_PRICE = b'step_price'
    STEP_COSTS = b'step_costs'
    MAX_STEP_LIMITS = b'max_step_limits'

    REVISION_CODE = b'revision_code'
    REVISION_NAME = b'revision_name'

    SCORE_BLACK_LIST = b'score_black_list'
    IMPORT_WHITE_LIST = b'import_white_list'

    IREP = b'irep'

    @classmethod
    def gs_migration_type_list(cls) -> list:
        return [
            cls.SERVICE_CONFIG,
            cls.STEP_PRICE,
            cls.STEP_COSTS,
            cls.MAX_STEP_LIMITS,
            cls.REVISION_CODE,
            cls.REVISION_NAME,
            cls.SCORE_BLACK_LIST,
            cls.IMPORT_WHITE_LIST,
        ]

    @classmethod
    def gs_migration_count(cls) -> int:
        return len(cls.gs_migration_type_list())


@unique
class IconScoreContextType(IntEnum):
    # Write data to db directly
    DIRECT = 0
    # Record data to cache and after confirming the block, write them to db
    INVOKE = 1
    # Record data to cache for estimation of steps, discard cache after estimation.
    ESTIMATION = 2
    # Not possible to write data to db
    QUERY = 3


@unique
class IconScoreFuncType(IntEnum):
    # ReadOnly function
    READONLY = 0
    # Writable function
    WRITABLE = 1


ENABLE_THREAD_FLAG = EnableThreadFlag.INVOKE | EnableThreadFlag.QUERY | EnableThreadFlag.VALIDATE


class DeployType(IntEnum):
    INSTALL = 0
    UPDATE = 1


class DeployState(IntEnum):
    INACTIVE = 0
    ACTIVE = 1


# 0xb9eeb235f715b166cf4b91ffcf8cc48a81913896086d30104ffc0cf47eed1cbd
INVALID_CLAIM_TX = [
    b'\xb9\xee\xb25\xf7\x15\xb1f\xcfK\x91\xff\xcf\x8c\xc4\x8a\x81\x918\x96\x08m0\x10O\xfc\x0c\xf4~\xed\x1c\xbd'
]

HASH_TYPE_TABLE = [
    "blockHash",
    "txHash",
    "prevBlockHash",
    "rootHash"
]

PREP_MAIN_PREPS = 22
PREP_MAIN_AND_SUB_PREPS = 100
PREP_REGISTRATION_FEE = 2_000 * ICX_IN_LOOP

IISS_MAX_REWARD_RATE = 10_000
IISS_MIN_IREP = 10_000 * ICX_IN_LOOP
IISS_MAX_IREP_PERCENTAGE = 14
IISS_INITIAL_IREP = 50_000 * ICX_IN_LOOP

# 24 hours * 60 minutes * 60 seconds / 2 - 80 <- for PRep terms
TERM_PERIOD = 24 * 60 * 60 // 2 - 80

# 24 hours * 60 minutes * 60 seconds / 2
IISS_DAY_BLOCK = 24 * 60 * 60 // 2
IISS_MONTH_BLOCK = IISS_DAY_BLOCK * 30
IISS_MONTH = 12
IISS_ANNUAL_BLOCK = IISS_MONTH_BLOCK * IISS_MONTH

UNSTAKE_SLOT_MAX = 1_000

PERCENTAGE_FOR_BETA_2 = 100

ISCORE_EXCHANGE_RATE = 1_000

PENALTY_GRACE_PERIOD = IISS_DAY_BLOCK * 2

LOW_PRODUCTIVITY_PENALTY_THRESHOLD = 85  # Unit: Percent
BLOCK_VALIDATION_PENALTY_THRESHOLD = 660  # Unit: Blocks

BASE_TRANSACTION_VERSION = 3

PREP_PENALTY_SIGNATURE = "PenaltyImposed(Address,int,int)"

BACKUP_FILES = 10

BLOCK_INVOKE_TIMEOUT_S = 15


class RCStatus(IntEnum):
    NOT_READY = 0
    READY = 1


class RCCalculateResult(IntEnum):
    SUCCESS = 0
    FAIL = 1
    IN_PROGRESS = 2
    INVALID_BLOCK_HEIGHT = 3


class PRepStatus(Enum):
    ACTIVE = 0
    UNREGISTERED = auto()
    DISQUALIFIED = auto()


class PenaltyReason(Enum):
    NONE = 0
    # disqualified
    PREP_DISQUALIFICATION = auto()
    LOW_PRODUCTIVITY = auto()
    # suspended
    BLOCK_VALIDATION = auto()


class PRepGrade(Enum):
    MAIN = 0
    SUB = 1
    CANDIDATE = 2


class PRepResultState(Enum):
    NORMAL = 0
    IN_TERM_UPDATED = 1


class BlockVoteStatus(Enum):
    NONE = 0
    TRUE = 1
    FALSE = 2


class PRepFlag(Flag):
    """Setting flags to True means that PRep fields specified by the flags has been changed

    """
    NONE = 0
    STATUS = auto()
    NAME = auto()
    COUNTRY = auto()
    CITY = auto()
    EMAIL = auto()
    WEBSITE = auto()
    DETAILS = auto()
    P2P_ENDPOINT = auto()
    PENALTY = auto()
    GRADE = auto()
    STAKE = auto()
    DELEGATED = auto()
    LAST_GENERATE_BLOCK_HEIGHT = auto()
    TOTAL_BLOCKS = auto()
    VALIDATED_BLOCKS = auto()
    UNVALIDATED_SEQUENCE_BLOCKS = auto()
    IREP = auto()  # irep, irep_block_height
    IREP_BLOCK_HEIGHT = auto()
    NODE_ADDRESS = auto()

    BLOCK_STATISTICS = TOTAL_BLOCKS | VALIDATED_BLOCKS | UNVALIDATED_SEQUENCE_BLOCKS
    ALL = 0xFFFFFFFF


class PRepContainerFlag(Flag):
    NONE = 0
    DIRTY = auto()


class TermFlag(Flag):
    NONE = 0
    MAIN_PREPS = auto()
    SUB_PREPS = auto()
    MAIN_PREP_P2P_ENDPOINT = auto()
    MAIN_PREP_NODE_ADDRESS = auto()

    ALL = 0xFFFFFFFF


class RevisionChangedFlag(Flag):
    # Empty
    NONE = 0x0
    # Set when STEP price changed on the block
    # STEP_PRICE_CHANGED = 0x10
    # Set when STEP costs changed on the block
    # STEP_COST_CHANGED = 0x20
    # Set when Max STEP limits changed on the block
    # STEP_MAX_LIMIT_CHANGED = 0x40
    # STEP changed flag mask
    # STEP_ALL_CHANGED = 0xf0

    # CHANGE REVISION
    GENESIS_IISS_CALC = 0x100
    IISS_CALC = 0x200
    DECENTRALIZATION = 0x400


class RPCMethod:
    ICX_GET_BALANCE = 'icx_getBalance'
    ICX_GET_TOTAL_SUPPLY = 'icx_getTotalSupply'
    ICX_GET_SCORE_API = 'icx_getScoreApi'
    ISE_GET_STATUS = 'ise_getStatus'
    ICX_CALL = 'icx_call'
    ICX_SEND_TRANSACTION = 'icx_sendTransaction'
    DEBUG_ESTIMATE_STEP = "debug_estimateStep"



# -*- coding: utf-8 -*-

# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities

Functions and classes in this module don't have any external dependencies.
"""

import hashlib
import json
import re
from collections import namedtuple
from enum import Flag
from typing import Any, Union, Optional
import inspect

from iconcommons import Logger

from ..icon_constant import BUILTIN_SCORE_ADDRESS_MAPPER, DATA_BYTE_ORDER, ICX_IN_LOOP


def int_to_bytes(n: int) -> bytes:
    length = byte_length_of_int(n)
    return n.to_bytes(length, byteorder=DATA_BYTE_ORDER, signed=True)


def bytes_to_int(v: bytes) -> int:
    return int.from_bytes(v, byteorder=DATA_BYTE_ORDER, signed=True)


def byte_length_of_int(n: int):
    if n < 0:
        # adds 1 because `bit_length()` always returns a bit length of absolute-value of `n`
        n += 1
    return (n.bit_length() + 8) // 8


def bytes_to_hex(data: Optional[bytes], prefix: str = "0x") -> str:
    if not isinstance(data, bytes):
        return "None"

    return f"{prefix}{data.hex()}"


def icx_to_loop(icx: int) -> int:
    return icx * ICX_IN_LOOP


def is_lowercase_hex_string(value: str) -> bool:
    """Check whether value is hexadecimal format or not

    :param value: text
    :return: True(lowercase hexadecimal) otherwise False
    """

    try:
        result = re.match('[0-9a-f]+', value)
        return len(result.group(0)) == len(value)
    except:
        pass

    return False


def sha3_256(data: bytes) -> bytes:
    return hashlib.sha3_256(data).digest()


def to_camel_case(snake_str: str) -> str:
    str_array = snake_str.split('_')
    return str_array[0] + ''.join(sub.title() for sub in str_array[1:])


def check_error_response(result: Any):
    return isinstance(result, dict) and result.get('error')


def get_main_type_from_annotations_type(annotations_type: type) -> type:
    main_type = None

    if hasattr(annotations_type, '__origin__') and annotations_type.__origin__ is not Union:
        return annotations_type.__origin__

    # in python 3.7, _subs_tree method has excluded.
    if hasattr(annotations_type, '__args__'):
        annotations = annotations_type.__args__
        for annotation_type in annotations:
            if annotation_type is not None:
                main_type = annotation_type
                break
    else:
        main_type = annotations_type
    return main_type


def is_builtin_score(score_address: str) -> bool:
    return score_address in BUILTIN_SCORE_ADDRESS_MAPPER.values()


def is_all_flag_on(src_flags: Flag, flag: Flag) -> bool:
    return src_flags & flag == flag


def is_any_flag_on(src_flags: Flag, flag: Flag) -> bool:
    return bool(src_flags & flag)


def set_flag(src_flags: Flag, flag: Flag, on: bool) -> Flag:
    if on:
        src_flags |= flag
    else:
        src_flags &= ~flag

    return src_flags


ContextEngine = namedtuple("engine", "deploy fee icx iiss prep issue inv")
ContextStorage = namedtuple("storage", "deploy fee icx iiss prep issue meta rc inv")


def print_log_with_level(log_level: str, msg: str, tag: str):
    logging_function = getattr(Logger, log_level)
    logging_function(msg, tag)


class BytesToHexJSONEncoder(json.JSONEncoder):
    """Used to make logging message more readable

    """
    def default(self, obj):
        if isinstance(obj, bytes):
            return bytes_to_hex(obj)

        return json.JSONEncoder.default(self, obj)



# -*- coding: utf-8 -*-

# Copyright 2019 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc


class HashOriginGenerator(abc.ABC):
    version = None

    @abc.abstractmethod
    def generate(self, origin_data: dict) -> str:
        pass


class HashOriginGeneratorV1(HashOriginGenerator):
    version = 1

    _translator = str.maketrans({
        "\\": "\\\\",
        "{": "\\{",
        "}": "\\}",
        "[": "\\[",
        "]": "\\]",
        ".": "\\."
    })

    def generate(self, json_data: dict):

        def encode(data):
            if isinstance(data, dict):
                return encode_dict(data)
            elif isinstance(data, list):
                return encode_list(data)
            else:
                return escape(data)

        def encode_dict(data: dict):
            result = ".".join(_encode_dict(data))
            return "{" + result + "}"

        def _encode_dict(data: dict):
            for key in sorted(data.keys()):
                yield key
                yield encode(data[key])

        def encode_list(data: list):
            result = ".".join(_encode_list(data))
            return f"[" + result + "]"

        def _encode_list(data: list):
            for item in data:
                yield encode(item)

        def escape(data):
            if data is None:
                return "\\0"

            data = str(data)
            return data.translate(self._translator)

        return ".".join(_encode_dict(json_data))



# Copyright 2019 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from copy import deepcopy
from typing import TYPE_CHECKING, Any, Optional, List, Dict, Tuple

from iconcommons.logger import Logger

from .data import Term
from .data.prep import PRep, PRepDictType
from .data.prep_container import PRepContainer
from .penalty_imposer import PenaltyImposer
from .validator import validate_prep_data, validate_irep
from ..base.ComponentBase import EngineBase
from ..base.address import Address, SYSTEM_SCORE_ADDRESS
from ..base.exception import (
    AccessDeniedException, InvalidParamsException, MethodNotFoundException, ServiceNotReadyException,
    InvalidRequestException
)
from ..base.type_converter_templates import ConstantKeys
from ..icon_constant import PRepGrade, PRepResultState, PRepStatus, ROLLBACK_LOG_TAG
from ..icon_constant import Revision, IISS_MIN_IREP, PREP_PENALTY_SIGNATURE, \
    PenaltyReason, TermFlag, IconScoreContextType
from ..iconscore.icon_score_context import IconScoreContext
from ..iconscore.icon_score_event_log import EventLogEmitter
from ..iconscore.icon_score_step import StepType
from ..icx.icx_account import Account
from ..icx.storage import Intent
from ..iiss.listener import EngineListener as IISSEngineListener
from ..iiss.reward_calc import RewardCalcDataCreator
from ..prep.prep_address_converter import PRepAddressConverter

if TYPE_CHECKING:
    from ..iiss.reward_calc.msg_data import PRepRegisterTx, PRepUnregisterTx, TxData
    from ..icx import IcxStorage
    from ..precommit_data_manager import PrecommitData

_TAG = "PREP"


class Method:
    REGISTER = 'registerPRep'
    UNREGISTER = 'unregisterPRep'
    SET_PREP = 'setPRep'
    SET_GOVERNANCE_VARIABLES = 'setGovernanceVariables'
    GET_PREP = 'getPRep'
    GET_MAIN_PREPS = 'getMainPReps'
    GET_SUB_PREPS = 'getSubPReps'
    GET_PREPS = 'getPReps'
    GET_PREP_TERM = 'getPRepTerm'
    GET_INACTIVE_PREPS = 'getInactivePReps'


class Engine(EngineBase, IISSEngineListener):
    """PRepEngine class

    Roles:
    * Manages term and preps
    * Handles P-Rep related JSON-RPC API requests
    """

    INVOKE_METHOD_TABLE = [
        Method.REGISTER,
        Method.UNREGISTER,
        Method.SET_PREP,
        Method.SET_GOVERNANCE_VARIABLES,
    ]
    QUERY_METHOD_TABLE = [
        Method.GET_PREP,
        Method.GET_MAIN_PREPS,
        Method.GET_SUB_PREPS,
        Method.GET_PREPS,
        Method.GET_PREP_TERM,
        Method.GET_INACTIVE_PREPS,
    ]
    METHOD_TABLE = INVOKE_METHOD_TABLE + QUERY_METHOD_TABLE

    def __init__(self):
        super().__init__()
        Logger.debug(tag=_TAG, msg="PRepEngine.__init__() start")

        self._invoke_handlers: dict = {
            Method.REGISTER: self.handle_register_prep,
            Method.UNREGISTER: self.handle_unregister_prep,
            Method.SET_PREP: self.handle_set_prep,
            Method.SET_GOVERNANCE_VARIABLES: self.handle_set_governance_variables,
            Method.GET_PREP: self.handle_get_prep,
            Method.GET_MAIN_PREPS: self.handle_get_main_preps,
            Method.GET_SUB_PREPS: self.handle_get_sub_preps,
            Method.GET_PREPS: self.handle_get_preps,
            Method.GET_PREP_TERM: self.handle_get_prep_term,
            Method.GET_INACTIVE_PREPS: self.handle_get_inactive_preps
        }

        self._query_handler: dict = {
            Method.GET_PREP: self.handle_get_prep,
            Method.GET_MAIN_PREPS: self.handle_get_main_preps,
            Method.GET_SUB_PREPS: self.handle_get_sub_preps,
            Method.GET_PREPS: self.handle_get_preps,
            Method.GET_PREP_TERM: self.handle_get_prep_term,
            Method.GET_INACTIVE_PREPS: self.handle_get_inactive_preps
        }

        self.preps: 'PRepContainer' = PRepContainer()
        # self.term should be None before decentralization
        self.term: Optional['Term'] = None
        self._initial_irep: Optional[int] = None
        self._penalty_imposer: Optional['PenaltyImposer'] = None

        self.prep_address_converter: 'PRepAddressConverter' = None

        Logger.debug(tag=_TAG, msg="PRepEngine.__init__() end")

    def open(self,
             context: 'IconScoreContext',
             term_period: int,
             irep: int,
             penalty_grace_period: int,
             low_productivity_penalty_threshold: int,
             block_validation_penalty_threshold: int):

        # This logic doesn't need to save to DB yet
        self._init_penalty_imposer(penalty_grace_period,
                                   low_productivity_penalty_threshold,
                                   block_validation_penalty_threshold)

        self.prep_address_converter: 'PRepAddressConverter' = context.storage.meta.get_prep_address_converter(context)

        self.preps = self._load_preps(context)
        self.term = self._load_term(context)
        self._initial_irep = irep

        context.engine.iiss.add_listener(self)

    def _init_penalty_imposer(self,
                              penalty_grace_period: int,
                              low_productivity_penalty_threshold: int,
                              block_validation_penalty_threshold: int):
        """Initialize PenaltyImposer

        :param penalty_grace_period:
        :param low_productivity_penalty_threshold:
        :param block_validation_penalty_threshold:
        :return:
        """
        self._penalty_imposer = PenaltyImposer(penalty_grace_period,
                                               low_productivity_penalty_threshold,
                                               block_validation_penalty_threshold)

    def _load_preps(self, context: 'IconScoreContext') -> 'PRepContainer':
        """Load preps from state db

        :return: new prep container instance
        """
        icx_storage: 'IcxStorage' = context.storage.icx
        preps = PRepContainer()

        for prep in context.storage.prep.get_prep_iterator():

            if prep.status == PRepStatus.ACTIVE:
                self.prep_address_converter.add_node_address(node=prep.node_address, prep=prep.address)

            account: 'Account' = icx_storage.get_account(context, prep.address, Intent.ALL)

            prep.stake = account.stake
            prep.delegated = account.delegated_amount

            preps.add(prep)

        preps.freeze()
        return preps

    @classmethod
    def _load_term(cls, context: 'IconScoreContext') -> Optional['Term']:
        term = context.storage.prep.get_term(context)
        if term:
            term.freeze()

        return term

    def close(self):
        IconScoreContext.engine.iiss.remove_listener(self)

    @classmethod
    def check_method(cls, method: str) -> bool:
        return method in cls.METHOD_TABLE

    @classmethod
    def check_invoke_method(cls, method: str) -> bool:
        return method in cls.INVOKE_METHOD_TABLE

    @classmethod
    def check_query_method(cls, method: str) -> bool:
        return method in cls.QUERY_METHOD_TABLE

    def invoke(self, context: 'IconScoreContext', method: str, params: dict):
        if context.revision < Revision.IISS.value:
            context.step_counter.apply_step(StepType.CONTRACT_CALL, 1)
            raise InvalidParamsException(f"Method Not Found: {method}")
        handler: callable = self._invoke_handlers[method]
        handler(context, **params)

    def query(self, context: 'IconScoreContext', method: str, params: dict) -> Any:
        if context.revision < Revision.SYSTEM_SCORE_ENABLED.value and \
                context.type == IconScoreContextType.INVOKE:
            raise InvalidRequestException(f"Do not call readonly method '{method}' with 'icx_sendTransaction'")
        handler: callable = self._query_handler[method]
        ret = handler(context, **params)
        return ret

    def commit(self, _context: 'IconScoreContext', precommit_data: 'PrecommitData'):
        """If the current P-Rep term is over, update term with new information
        which has P-Rep list(address, delegated amount), start height, end height, irep

        :param _context:
        :param precommit_data:
        :return:
        """
        # Updated every block
        self.preps = precommit_data.preps

        # Exchange a term instance for some reasons:
        # - penalty for elected P-Reps(main, sub)
        # - A new term is started
        if precommit_data.term is not None:
            self.term: 'Term' = precommit_data.term

        self.prep_address_converter: 'PRepAddressConverter' = precommit_data.prep_address_converter

    def rollback(self, context: 'IconScoreContext', _block_height: int, _block_hash: bytes):
        """After rollback is called, the state of prep_engine is reverted to that of a given block

        :param context:
        :param _block_height: the height of the block to go back
        :param _block_hash:
        :return:
        """
        Logger.info(tag=ROLLBACK_LOG_TAG, msg="rollback() start")

        self.prep_address_converter: 'PRepAddressConverter' = context.storage.meta.get_prep_address_converter(context)

        self.preps = self._load_preps(context)
        self.term = self._load_term(context)

        Logger.info(tag=ROLLBACK_LOG_TAG, msg=f"rollback() end: {self.term}")

    def on_block_invoked(
            self,
            context: 'IconScoreContext',
            is_decentralization_started: bool) -> Tuple[Optional[dict], Optional['Term']]:
        """Called on IconServiceEngine._after_transaction_process()

        1. Adjust the grade for invalid P-Reps
        2. Update elected P-Reps in this term

        :param context:
        :param is_decentralization_started:
            True: Decentralization will begin at the next block
        :return:
        """

        if is_decentralization_started or self._is_term_ended(context):
            # The current P-Rep term is over. Prepare the next P-Rep term
            next_preps, new_term = self._on_term_ended(context)
        elif context.is_decentralized():
            # In-term P-Rep replacement
            next_preps, new_term = self._on_term_updated(context)
        else:
            next_preps, new_term = None, None

        if new_term:
            self._update_prep_grades(context, context.preps, self.term, new_term)
            context.storage.prep.put_term(context, new_term)

        return next_preps, new_term

    def _is_term_ended(self, context: 'IconScoreContext') -> bool:
        if self.term is None:
            return False

        return context.block.height == self.term.end_block_height

    def _on_term_ended(self, context: 'IconScoreContext') -> Tuple[dict, 'Term']:
        """Called in IconServiceEngine.invoke() every time when a term is ended

        Update P-Rep grades according to PRep.delegated
        """
        self._put_last_term_info(context, self.term)

        if self.term:
            main_preps: List['Address'] = [prep.address for prep in self.term.main_preps]
        else:
            # first term
            new_preps: List['PRep'] = context.preps.get_preps(start_index=0,
                                                              size=context.main_prep_count)
            main_preps: List['Address'] = [prep.address for prep in new_preps]

        context.storage.meta.put_last_main_preps(context, main_preps)

        # All block validation penalties are released
        self._reset_block_validation_penalty(context)

        # Create a term with context.preps whose grades are up-to-date
        new_term: 'Term' = self._create_next_term(context, self.term)
        next_preps: dict = self._get_updated_main_preps(
            context=context,
            term=new_term,
            state=PRepResultState.NORMAL
        )

        Logger.debug(tag=_TAG, msg=f"{new_term}")

        return next_preps, new_term

    @classmethod
    def _put_last_term_info(cls, context: 'IconScoreContext', term: 'Term'):

        _, last_calc_end = context.storage.meta.get_last_term_info(context)
        if last_calc_end > 0:
            start: int = term.start_block_height
            end: int = term.end_block_height
        else:
            # first
            start: int = -1
            end: int = context.block.height

        context.storage.meta.put_last_term_info(context, start, end)

    def _on_term_updated(self, context: 'IconScoreContext') -> Tuple[Optional[dict], Optional['Term']]:
        """Update term with invalid elected P-Rep list during this term
        (In-term P-Rep replacement)

        We have to consider 5 cases below:
        1. No invalid elected P-Rep
            - Nothing to do
        2. Only main P-Reps are invalidated
            - Send a new main P-Rep list to loopchain
            - Save the new term to DB
        3. Only sub P-Reps are invalidated
            - Save the new term to DB
        4. Both of them are invalidated
            - Send new main P-Rep list to loopchain
            - Save the new term to DB
        5. p2pEndpoint of a Main P-Rep is updated
            - Send new main P-Rep list to loopchain
            - No need to save the new term to DB

        :param context:
        :return:
        """

        main_preps: List['Address'] = [prep.address for prep in self.term.main_preps]
        context.storage.meta.put_last_main_preps(context, main_preps)

        new_term = context.term
        if not new_term.is_dirty():
            return None, None

        if bool(new_term.flags & (TermFlag.MAIN_PREPS |
                                  TermFlag.MAIN_PREP_P2P_ENDPOINT |
                                  TermFlag.MAIN_PREP_NODE_ADDRESS)):
            next_preps = self._get_updated_main_preps(
                context=context,
                term=new_term,
                state=PRepResultState.IN_TERM_UPDATED
            )
        else:
            next_preps = None

        return next_preps, new_term

    @classmethod
    def _update_prep_grades(cls,
                            context: 'IconScoreContext',
                            new_preps: 'PRepContainer',
                            old_term: Optional['Term'],
                            new_term: 'Term'):
        """Update the grades of P-Reps every time when a block is invoked

        Do NOT change any properties of P-Reps after this method is called in this block

        :param context:
        :return:
        """
        Logger.debug(tag=_TAG, msg="_update_prep_grades() start")

        # 0: old grade, 1: new grade
        _OLD, _NEW = 0, 1
        prep_grades: Dict['Address', Tuple['PRepGrade', 'PRepGrade']] = {}

        if old_term:
            for prep_snapshot in old_term.main_preps:
                prep_grades[prep_snapshot.address] = (PRepGrade.MAIN, PRepGrade.CANDIDATE)

            for prep_snapshot in old_term.sub_preps:
                prep_grades[prep_snapshot.address] = (PRepGrade.SUB, PRepGrade.CANDIDATE)

        # Remove the P-Reps which preserve the same grade in the next term from prep_grades dict
        main_prep_count: int = len(new_term.main_preps)
        for i, prep_snapshot in enumerate(new_term.preps):
            prep_address: 'Address' = prep_snapshot.address
            grades: tuple = prep_grades.get(prep_address, (PRepGrade.CANDIDATE, PRepGrade.CANDIDATE))

            old_grade: 'PRepGrade' = grades[_OLD]
            new_grade: 'PRepGrade' = PRepGrade.MAIN if i < main_prep_count else PRepGrade.SUB

            if old_grade == new_grade:
                del prep_grades[prep_address]
            else:
                prep_grades[prep_address] = (old_grade, new_grade)

        # Update the grades of P-Reps for the next term
        # CAUTION: DO NOT use context.put_dirty_prep() here
        for address, grades in prep_grades.items():
            prep: 'PRep' = new_preps.get_by_address(address)
            assert prep is not None

            if prep.grade == grades[_NEW]:
                continue

            prep = prep.copy()
            prep.grade = grades[_NEW]
            new_preps.replace(prep)
            context.storage.prep.put_prep(context, prep)

            Logger.info(tag=_TAG,
                        msg=f"P-Rep grade changed: {address} {grades[_OLD]} -> {grades[_NEW]}")

        Logger.debug(tag=_TAG, msg="_update_prep_grades() end")

    @classmethod
    def _reset_block_validation_penalty(cls, context: 'IconScoreContext'):
        """Reset block validation penalty in the end of every term

        :param context:
        :return:
        """

        for prep in context.preps:
            if prep.penalty == PenaltyReason.BLOCK_VALIDATION and prep.status == PRepStatus.ACTIVE:
                dirty_prep = context.get_prep(prep.address, mutable=True)
                dirty_prep.reset_block_validation_penalty()
                context.put_dirty_prep(dirty_prep)

        context.update_dirty_prep_batch()

    def handle_register_prep(self, context: 'IconScoreContext', **kwargs):
        """Register a P-Rep

        Roles
        * Update preps in context
        * Update stateDB
        * Update rcDB
        """
        if context.msg.sender.is_contract:
            raise AccessDeniedException(f"SCORE is not allowed.")

        icx_storage: 'IcxStorage' = context.storage.icx

        address: 'Address' = context.tx.origin
        if context.preps.contains(address, active_prep_only=False):
            raise InvalidParamsException(f"P-Rep already exists: {address}")

        # Check Prep registration fee
        value = context.msg.value
        if value != context.storage.prep.prep_registration_fee:
            raise InvalidParamsException(f"Invalid prep registration fee. "
                                         f"Registration Fee Must be {context.storage.prep.prep_registration_fee} "
                                         f"not {value}")

        if context.revision < Revision.DIVIDE_NODE_ADDRESS.value:
            self._remove_node_address_from_params(params=kwargs)

        validate_prep_data(context=context,
                           prep_address=address,
                           tx_data=kwargs)

        account: 'Account' = icx_storage.get_account(context, address, Intent.STAKE | Intent.DELEGATED)

        # Create a PRep object and assign delegated amount from account to prep
        # prep.irep is set to IISS_MIN_IREP by default
        dirty_prep = PRep.from_dict(address, kwargs, context.block.height, context.tx.index)
        dirty_prep.stake = account.stake
        dirty_prep.delegated = account.delegated_amount

        # Set an initial value to irep of a P-Rep on registerPRep
        if context.is_decentralized():
            dirty_prep.set_irep(self.term.irep, context.block.height)
        else:
            dirty_prep.set_irep(self._initial_irep, context.block.height)

        # Update preps in context
        context.put_dirty_prep(dirty_prep)

        # Update rcDB
        self._put_reg_prep_in_rc_db(context, address)

        # Burn Prep registration fee
        context.engine.issue.burn(context, address, value)

        # EventLog
        EventLogEmitter.emit_event_log(
            context,
            score_address=SYSTEM_SCORE_ADDRESS,
            event_signature="PRepRegistered(Address)",
            arguments=[address],
            indexed_args_count=0
        )

    @classmethod
    def _remove_node_address_from_params(cls, params: dict):
        """Just for backward compatibility with the previous version

        :param params: parameters of registerPRep or setPRep
        """
        if ConstantKeys.NODE_ADDRESS in params:
            del params[ConstantKeys.NODE_ADDRESS]

    @classmethod
    def _put_reg_prep_in_rc_db(cls, context: 'IconScoreContext', address: 'Address'):
        """Put a newly registered P-Rep in RewardCalcDatabase

        :param context:
        :param address: The address of P-Rep
        :return:
        """

        rc_tx_batch: list = context.rc_tx_batch
        block_height: int = context.block.height

        tx: 'PRepRegisterTx' = RewardCalcDataCreator.create_tx_prep_reg()
        iiss_tx_data: 'TxData' = RewardCalcDataCreator.create_tx(address, block_height, tx)
        context.storage.rc.put(rc_tx_batch, iiss_tx_data)

    @classmethod
    def _get_updated_main_preps(cls,
                                context: 'IconScoreContext',
                                term: 'Term',
                                state: 'PRepResultState') -> Optional[dict]:
        """Returns preps which will run as main preps during the next term in dict format

        :return:
        """
        updated_main_preps: Optional[dict] = \
            cls.get_main_preps_in_dict(context, term)

        if updated_main_preps:
            updated_main_preps['irep'] = term.irep
            updated_main_preps['state'] = state.value

        return updated_main_preps

    @classmethod
    def get_main_preps_in_dict(cls,
                               context: 'IconScoreContext',
                               term: 'Term') -> Optional[dict]:
        if len(term.main_preps) == 0:
            Logger.warning(tag="PREP", msg="No P-Rep candidates")
            return None

        prep_as_dict: dict = {}
        preps_as_list: list = []

        for prep_snapshot in term.main_preps:
            prep: 'PRep' = context.get_prep(prep_snapshot.address)
            preps_as_list.append({
                ConstantKeys.PREP_ID: prep.node_address,
                ConstantKeys.P2P_ENDPOINT: prep.p2p_endpoint
            })

        prep_as_dict["preps"] = preps_as_list
        prep_as_dict["rootHash"]: bytes = term.root_hash

        return prep_as_dict

    @classmethod
    def _create_next_term(cls,
                          context: 'IconScoreContext',
                          prev_term: Optional['Term']) -> 'Term':
        """Create the next term instance at the end of the current term

        :param prev_term:
        :param context:
        :return:
        """
        new_preps: List['PRep'] = context.preps.get_preps(
            start_index=0, size=context.main_and_sub_prep_count)

        sequence = 0 if prev_term is None else prev_term.sequence + 1
        start_block_height = context.block.height + 1
        if prev_term:
            assert start_block_height == prev_term.end_block_height + 1

        # The current P-Rep term is over. Prepare the next P-Rep term
        if context.revision < Revision.SET_IREP_VIA_NETWORK_PROPOSAL.value:
            irep: int = cls._calculate_weighted_average_of_irep(new_preps[:context.main_prep_count])
        else:
            irep: int = context.inv_container.irep

        term = Term(
            sequence,
            start_block_height,
            context.term_period,
            irep,
            context.total_supply,
            context.preps.total_delegated
        )

        term.set_preps(new_preps, context.main_prep_count, context.main_and_sub_prep_count)

        return term

    @classmethod
    def _calculate_weighted_average_of_irep(cls, new_main_preps: List['PRep']) -> int:
        total_delegated = 0  # total delegated of top 22 preps
        total_weighted_irep = 0

        for prep in new_main_preps:
            total_weighted_irep += prep.irep * prep.delegated
            total_delegated += prep.delegated

        return total_weighted_irep // total_delegated if total_delegated > 0 else IISS_MIN_IREP

    def handle_get_prep(self, context: 'IconScoreContext', address: 'Address') -> dict:
        """Returns the details of a P-Rep including information on registration, delegation and statistics

        :return: the response for getPRep JSON-RPC request
        """
        prep: 'PRep' = self.preps.get_by_address(address)
        if prep is None:
            raise InvalidParamsException(f"P-Rep not found: {address}")

        response: dict = prep.to_dict(PRepDictType.FULL)
        return response

    @classmethod
    def handle_set_prep(cls, context: 'IconScoreContext', **kwargs):
        """Update a P-Rep registration information

        :param context:
        :param kwargs:
        :return:
        """
        if context.msg.sender.is_contract:
            raise AccessDeniedException(f"SCORE is not allowed.")

        address: 'Address' = context.tx.origin

        dirty_prep: Optional['PRep'] = context.get_prep(address, mutable=True)
        if dirty_prep is None:
            raise InvalidParamsException(f"P-Rep not found: {address}")

        params: dict = deepcopy(kwargs)

        if context.revision < Revision.DIVIDE_NODE_ADDRESS.value:
            cls._remove_node_address_from_params(params=params)

        validate_prep_data(context=context,
                           prep_address=address,
                           tx_data=params,
                           set_prep=True)

        if ConstantKeys.P2P_ENDPOINT in params:
            p2p_endpoint: str = params[ConstantKeys.P2P_ENDPOINT]
            del params[ConstantKeys.P2P_ENDPOINT]
            params["p2p_endpoint"] = p2p_endpoint

        if ConstantKeys.NODE_ADDRESS in params:
            node_address: 'Address' = params[ConstantKeys.NODE_ADDRESS]
            del params[ConstantKeys.NODE_ADDRESS]
            params["node_address"] = node_address

        # EventLog
        EventLogEmitter.emit_event_log(
            context,
            score_address=SYSTEM_SCORE_ADDRESS,
            event_signature="PRepSet(Address)",
            arguments=[address],
            indexed_args_count=0
        )

        cls._validate_node_key_back_compatibillity_below_rev_9(context, kwargs)
        # Update registration info
        dirty_prep.set(**params)

        context.put_dirty_prep(dirty_prep)

    @classmethod
    def _validate_node_key_back_compatibillity_below_rev_9(cls, context: 'IconScoreContext', data: dict):
        if context.revision < Revision.DIVIDE_NODE_ADDRESS.value:
            if ConstantKeys.NODE_ADDRESS in data and \
                    data[ConstantKeys.NODE_ADDRESS] is not None:
                # For Backward compatibility
                raise TypeError("nodeAddress not Allowed")


    def handle_set_governance_variables(self, context: 'IconScoreContext', irep: int):
        """Handles setGovernanceVariables JSON-RPC API request
        """
        if context.msg.sender.is_contract:
            raise AccessDeniedException(f"SCORE is not allowed.")

        # This API is available after IISS decentralization is enabled.
        if context.revision < Revision.DECENTRALIZATION.value or self.term.sequence < 0:
            raise MethodNotFoundException("setGovernanceVariables is disabled")

        # This API is disabled after SET_IREP_VIA_NETWORK_PROPOSAL
        if context.revision >= Revision.SET_IREP_VIA_NETWORK_PROPOSAL.value:
            raise MethodNotFoundException("setGovernanceVariables is disabled")

        address: 'Address' = context.tx.origin

        dirty_prep: Optional['PRep'] = context.get_prep(address, mutable=True)
        if dirty_prep is None:
            raise InvalidParamsException(f"P-Rep not found: {address}")

        # Update incentive rep
        validate_irep(context, irep, dirty_prep)

        # EventLog
        EventLogEmitter.emit_event_log(
            context,
            score_address=SYSTEM_SCORE_ADDRESS,
            event_signature="GovernanceVariablesSet(Address,int)",
            arguments=[address, irep],
            indexed_args_count=1
        )

        # Update the changed properties of a P-Rep to stateDB
        # context.storage.prep.put_dirty_prep(context, prep)
        dirty_prep.set_irep(irep, context.block.height)
        context.put_dirty_prep(dirty_prep)

    def handle_unregister_prep(self, context: 'IconScoreContext'):
        """Unregister a P-Rep

        :param context:
        :return:
        """
        if context.msg.sender.is_contract:
            raise AccessDeniedException(f"SCORE is not allowed.")

        address: 'Address' = context.tx.origin

        dirty_prep: Optional['PRep'] = context.get_prep(address, mutable=True)
        if dirty_prep is None:
            raise InvalidParamsException(f"P-Rep not found: {address}")

        if dirty_prep.status != PRepStatus.ACTIVE:
            raise InvalidParamsException(f"Inactive P-Rep: {address}")

        dirty_prep.status = PRepStatus.UNREGISTERED
        dirty_prep.grade = PRepGrade.CANDIDATE
        context.put_dirty_prep(dirty_prep)

        # Update rcDB
        self._put_unreg_prep_for_iiss_db(context, address)

        # EventLog
        EventLogEmitter.emit_event_log(
            context,
            score_address=SYSTEM_SCORE_ADDRESS,
            event_signature="PRepUnregistered(Address)",
            arguments=[address],
            indexed_args_count=0
        )

    def impose_penalty(self, context: 'IconScoreContext'):
        """Impose penalties on main P-Reps every block
        Called on IconServiceEngine._process_base_transaction()

        :param context:
        """
        for snapshot in self.term.main_preps:
            # Get a up-to-date main prep from context
            prep: 'PRep' = context.get_prep(snapshot.address)
            assert isinstance(prep, PRep)

            self._penalty_imposer.run(context, prep, self._on_penalty_imposed)

    @classmethod
    def _on_penalty_imposed(cls,
                            context: 'IconScoreContext',
                            address: 'Address',
                            reason: 'PenaltyReason'):
        """Called on PenaltyImposer.run()

        Penalty: low productivity, block validation, disqualification

        :param context:
        :param address: The address of a Main P-PRep on this term
        :param reason: penalty reason
        :return:
        """
        dirty_prep: Optional['PRep'] = context.get_prep(address, mutable=True)
        assert isinstance(dirty_prep, PRep)
        assert dirty_prep.status == PRepStatus.ACTIVE

        if reason == PenaltyReason.BLOCK_VALIDATION:
            status: 'PRepStatus' = PRepStatus.ACTIVE
        else:
            status: 'PRepStatus' = PRepStatus.DISQUALIFIED
            # Update rcDB not to supply a reward for the inactive P-Rep
            cls._put_unreg_prep_for_iiss_db(context, address)

        dirty_prep.status = status
        dirty_prep.penalty = reason
        dirty_prep.grade = PRepGrade.CANDIDATE
        context.put_dirty_prep(dirty_prep)

        EventLogEmitter.emit_event_log(
            context,
            score_address=SYSTEM_SCORE_ADDRESS,
            event_signature=PREP_PENALTY_SIGNATURE,
            arguments=[
                dirty_prep.address,
                dirty_prep.status.value,
                dirty_prep.penalty.value],
            indexed_args_count=1)

    @classmethod
    def impose_prep_disqualified_penalty(
            cls, context: 'IconScoreContext', address: 'Address'):
        """Called on disqualification network proposal

        :param context:
        :param address:
        :return:
        """
        prep: 'PRep' = context.get_prep(address)
        if prep is None:
            raise InvalidParamsException(f"P-Rep not found: {address}")

        if prep.status != PRepStatus.ACTIVE:
            raise InvalidParamsException(f"Inactive P-Rep: {address}")

        cls._on_penalty_imposed(context, address, PenaltyReason.PREP_DISQUALIFICATION)

    @classmethod
    def _put_unreg_prep_for_iiss_db(cls, context: 'IconScoreContext', address: 'Address'):
        rc_tx_batch: list = context.rc_tx_batch
        block_height: int = context.block.height

        tx: 'PRepUnregisterTx' = RewardCalcDataCreator.create_tx_prep_unreg()
        iiss_tx_data: 'TxData' = RewardCalcDataCreator.create_tx(address, block_height, tx)
        context.storage.rc.put(rc_tx_batch, iiss_tx_data)

    def handle_get_main_preps(self, _context: 'IconScoreContext') -> dict:
        """Returns main P-Rep list in the current term
        """
        total_delegated = 0 if self.term is None else self.term.total_delegated
        prep_list: list = []

        if self.term:
            for snapshot in self.term.main_preps:
                item = {
                    "address": snapshot.address,
                    "delegated": snapshot.delegated
                }
                prep_list.append(item)

        return {
            "totalDelegated": total_delegated,
            "preps": prep_list
        }

    def handle_get_sub_preps(self, _context: 'IconScoreContext') -> dict:
        """Returns sub P-Rep list in the present term
        """
        total_delegated: int = 0 if self.term is None else self.term.total_delegated
        prep_list: list = []

        if self.term:
            for prep in self.term.sub_preps:
                item = {
                    "address": prep.address,
                    "delegated": prep.delegated
                }
                prep_list.append(item)
                total_delegated += prep.delegated

        return {
            "totalDelegated": total_delegated,
            "preps": prep_list
        }

    def handle_get_preps(self,
                         context: "IconScoreContext",
                         startRanking: Optional[int],
                         endRanking: Optional[int]) -> dict:
        """
        Returns P-Reps ranging in ranking from startRanking to endRanking

        P-Rep means all P-Reps including main P-Reps and sub P-Reps
        """
        preps: 'PRepContainer' = self.preps
        prep_count: int = preps.size(active_prep_only=True)
        prep_list: list = []

        if startRanking is None:
            startRanking = 1
        if endRanking is None:
            endRanking = prep_count

        if prep_count > 0:
            if not 1 <= startRanking <= endRanking:
                raise InvalidParamsException(
                    f"Invalid ranking: startRanking({startRanking}), endRanking({endRanking})")

            for i in range(startRanking - 1, endRanking):
                prep: 'PRep' = preps.get_by_index(i)
                prep_list.append(prep.to_dict(PRepDictType.FULL))

        return {
            "blockHeight": context.block.height,
            "startRanking": startRanking,
            "totalDelegated": preps.total_delegated,
            "totalStake": context.storage.iiss.get_total_stake(context),
            "preps": prep_list
        }

    def handle_get_prep_term(self, context: 'IconScoreContext') -> dict:
        """Provides the information on the current term
        """
        if self.term is None:
            raise ServiceNotReadyException("Term is not ready")

        preps_data = []

        # Collect Main and Sub P-Reps
        for prep_snapshot in self.term.preps:
            prep = self.preps.get_by_address(prep_snapshot.address)
            preps_data.append(prep.to_dict(PRepDictType.FULL))

        # Collect P-Reps which got penalized for consecutive 660 block validation failure
        def _func(node: 'PRep') -> bool:
            return node.penalty == PenaltyReason.BLOCK_VALIDATION and node.status == PRepStatus.ACTIVE

        # Sort preps in descending order by delegated
        preps_on_block_validation_penalty = \
            sorted(filter(_func, self.preps), key=lambda x: x.order())

        for prep in preps_on_block_validation_penalty:
            preps_data.append(prep.to_dict(PRepDictType.FULL))

        return {
            "blockHeight": context.block.height,
            "sequence": self.term.sequence,
            "startBlockHeight": self.term.start_block_height,
            "endBlockHeight": self.term.end_block_height,
            "totalSupply": self.term.total_supply,
            "totalDelegated": self.term.total_delegated,
            "irep": self.term.irep,
            "preps": preps_data
        }

    def handle_get_inactive_preps(self, context: 'IconScoreContext') -> dict:
        """Returns inactive P-Reps which is unregistered or receiving prep disqualification or low productivity penalty.
        """
        sorted_inactive_preps: List['PRep'] = \
            sorted(self.preps.get_inactive_preps(), key=lambda node: node.order())

        total_delegated = 0
        inactive_preps_data = []
        for prep in sorted_inactive_preps:
            inactive_preps_data.append(prep.to_dict(PRepDictType.FULL))
            total_delegated += prep.delegated

        return {
            "blockHeight": context.block.height,
            "totalDelegated": total_delegated,
            "preps": inactive_preps_data
        }

    # IISSEngineListener implementation ---------------------------
    def on_set_stake(self, context: 'IconScoreContext', account: 'Account'):
        """Called on IISSEngine.handle_set_stake()

        :param context:
        :param account:
        :return:
        """
        dirty_prep: 'PRep' = context.get_prep(account.address, mutable=True)
        if dirty_prep:
            dirty_prep.stake = account.stake
            context.put_dirty_prep(dirty_prep)

    def on_set_delegation(
            self, context: 'IconScoreContext', updated_accounts: List['Account']):
        """Called on IISSEngine.handle_set_delegation()

        :param context:
        :param updated_accounts:
        return:
        """
        if context.revision <= Revision.DECENTRALIZATION.value:
            # Although the following statement has a bug,
            # preserve it for state compatibility
            # max delegations (i.e. 10) * 2 + 1 is correct
            assert 0 <= len(updated_accounts) <= context.engine.iiss.get_max_delegations_by_revision(context) * 2

        for account in updated_accounts:
            assert isinstance(account, Account)
            address = account.address

            # If a delegated account is a P-Rep, then update its delegated amount
            dirty_prep: Optional['PRep'] = context.get_prep(address, mutable=True)
            if dirty_prep:
                dirty_prep.delegated = account.delegated_amount
                context.put_dirty_prep(dirty_prep)



# Copyright 2019 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING, Optional

from ..base.type_converter import TypeConverter

if TYPE_CHECKING:
    from ..iconscore.icon_score_context import IconScoreContext
    from ..base.block import Block
    from ..prep.data import Term


def get_main_preps(context: 'IconScoreContext', **_kwargs):
    term: 'Term' = context.engine.prep.term
    preps: Optional[dict] = None
    if context.is_decentralized():
        preps: Optional[dict] = \
            context.engine.prep.get_main_preps_in_dict(context, term)
    if preps is None:
        preps = {}

    block: 'Block' = context.storage.icx.last_block
    preps['blockHeight'] = hex(0) if block is None else hex(block.height)
    TypeConverter.convert_type_reverse(preps)
    result = {
        "result": preps
    }

    return result


inner_call_handler = {
    "ise_getPRepList": get_main_preps
}


def inner_call(context: 'IconScoreContext', request: dict):
    method = request['method']
    params = request.get("params", {})
    handler = inner_call_handler[method]
    return handler(context, **params)



# -*- coding: utf-8 -*-

# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
from typing import TYPE_CHECKING, Optional, Tuple

from iconservice.score_loader.icon_score_class_loader import IconScoreClassLoader
from .icon_score_mapper_object import IconScoreInfo
from .score_package_validator import ScorePackageValidator
from .utils import get_package_name_by_address_and_tx_hash, get_score_deploy_path
from ..base.address import Address
from ..base.address import SYSTEM_SCORE_ADDRESS
from ..base.exception import ScoreNotFoundException, AccessDeniedException, FatalException
from ..database.db import IconScoreDatabase
from ..database.factory import ContextDatabaseFactory
from ..icon_constant import IconScoreContextType, IconServiceFlag, DeployState, BUILTIN_SCORE_IMPORT_WHITE_LIST
from ..utils import is_builtin_score

if TYPE_CHECKING:
    from .icon_score_context import IconScoreContext
    from .icon_score_base import IconScoreBase
    from .icon_score_mapper import IconScoreMapper
    from ..deploy.storage import IconScoreDeployTXParams, IconScoreDeployInfo


class IconScoreContextUtil(object):
    """Contains the useful information to process user's jsonrpc request
    """

    @staticmethod
    def is_score_active(context: 'IconScoreContext', score_address: 'Address') -> bool:
        if not score_address.is_contract:
            return False

        deploy_info: 'IconScoreDeployInfo' = \
            context.storage.deploy.get_deploy_info(context, score_address)

        if deploy_info is None:
            return False

        return deploy_info.deploy_state == DeployState.ACTIVE

    @staticmethod
    def get_owner(context: 'IconScoreContext',
                  score_address: 'Address') -> Optional['Address']:
        deploy_info: 'IconScoreDeployInfo' =\
            context.storage.deploy.get_deploy_info(context, score_address)

        if deploy_info is None:
            return None

        return deploy_info.owner

    @staticmethod
    def deploy(context: 'IconScoreContext', tx_hash: bytes) -> None:
        context.engine.deploy.deploy(context, tx_hash)

    @staticmethod
    def get_builtin_score(context: 'IconScoreContext', address: 'Address') -> 'IconScoreBase':
        score = IconScoreContextUtil.get_icon_score(context, address)

        if score is None:
            raise ScoreNotFoundException(f'Builtin SCORE not found: {address}')

        return score

    @staticmethod
    def get_icon_score(context: 'IconScoreContext', address: 'Address') -> Optional['IconScoreBase']:
        """Create a SCORE instance from SCORE class and return it

        :param context:
        :param address:
        :return:
        """
        score_info: 'IconScoreInfo' = IconScoreContextUtil.get_score_info(context, address)
        if score_info is None:
            return None

        # Create a SCORE instance every time
        # to prevent consensus failure by using wrong member variables in SCORE
        return score_info.get_score(context.revision)

    @staticmethod
    def get_score_info(context: 'IconScoreContext', address: 'Address') -> Optional['IconScoreInfo']:
        """Returns the score_info associated with the currently active score

        If IconScoreInfo instance is not found,
        create it from SCORE class and current tx_hash of the deployed SCORE

        :param context:
        :param address:
        :return:
        """
        score_mapper: 'IconScoreMapper' = context.icon_score_mapper
        deploy_info: 'IconScoreDeployInfo' = IconScoreContextUtil.get_deploy_info(context, address)

        if deploy_info is None or deploy_info.deploy_state != DeployState.ACTIVE:
            return None

        score_info: 'IconScoreInfo' = None

        if context.type == IconScoreContextType.INVOKE:
            score_info = context.new_icon_score_mapper.get(address)

        if score_info is None:
            score_info = score_mapper.get(address)

        current_tx_hash: bytes = deploy_info.current_tx_hash

        if score_info is None:
            score_info: 'IconScoreInfo' =\
                IconScoreContextUtil.create_score_info(context, address, current_tx_hash)
            score_mapper[address] = score_info
        elif score_info.tx_hash != current_tx_hash:
            raise FatalException(
                f'scoreInfo.txHash(0x{score_info.tx_hash.hex()}) != txHash(0x{current_tx_hash.hex()})')

        return score_info

    @staticmethod
    def create_score_info(
            context: 'IconScoreContext', score_address: 'Address',
            tx_hash: bytes, score_db: 'IconScoreDatabase' = None) -> 'IconScoreInfo':

        score_class: type = IconScoreClassLoader.run(
            score_address, tx_hash, context.score_root_path)

        if score_db is None:
            context_db = ContextDatabaseFactory.create_by_address(score_address)
            score_db = IconScoreDatabase(score_address, context_db)

        # Cache a new IconScoreInfo instance
        return IconScoreInfo(score_class, score_db, tx_hash)

    @staticmethod
    def validate_score_package(context: 'IconScoreContext', address: 'Address', tx_hash: bytes) -> None:

        if not IconScoreContextUtil.is_service_flag_on(context, IconServiceFlag.SCORE_PACKAGE_VALIDATOR):
            return

        score_deploy_path: str = get_score_deploy_path(context.score_root_path, address, tx_hash)
        score_package_name: str = get_package_name_by_address_and_tx_hash(address, tx_hash)
        import_whitelist: dict = context.inv_container.import_white_list

        # add import white list for builtin SCORE
        if is_builtin_score(str(address)):
            import_whitelist.update(BUILTIN_SCORE_IMPORT_WHITE_LIST)

        ScorePackageValidator.execute(import_whitelist, score_deploy_path, score_package_name)

    @staticmethod
    def validate_score_blacklist(context: 'IconScoreContext', score_address: 'Address') -> None:
        """Prevent SCOREs in blacklist

        :param context:
        :param score_address:
        """
        if not score_address.is_contract:
            return
        if score_address == SYSTEM_SCORE_ADDRESS:
            return

        if score_address in context.inv_container.score_black_list:
            raise AccessDeniedException(f'SCORE in blacklist: {score_address}')

    @staticmethod
    def is_service_flag_on(context: 'IconScoreContext', flag: 'IconServiceFlag') -> bool:
        if context.inv_container:
            service_flag = context.inv_container.service_config
        else:
            service_flag = context.icon_service_flag
        return IconScoreContextUtil._is_flag_on(service_flag, flag)

    @staticmethod
    def _is_flag_on(src_flag: int, dst_flag: int) -> bool:
        return src_flag & dst_flag == dst_flag

    @staticmethod
    def get_tx_hashes_by_score_address(context: 'IconScoreContext',
                                       score_address: 'Address') -> Tuple[Optional[bytes], Optional[bytes]]:
        warnings.warn("legacy function don't use.", DeprecationWarning, stacklevel=2)
        return context.storage.deploy.get_tx_hashes_by_score_address(
            context, score_address)

    @staticmethod
    def get_score_address_by_tx_hash(context: 'IconScoreContext',
                                     tx_hash: bytes) -> Optional['Address']:
        warnings.warn("legacy function don't use.", DeprecationWarning, stacklevel=2)
        return context.storage.deploy.get_score_address_by_tx_hash(context, tx_hash)

    @staticmethod
    def get_deploy_tx_params(context: 'IconScoreContext', tx_hash: bytes) -> Optional['IconScoreDeployTXParams']:
        return context.storage.deploy.get_deploy_tx_params(context, tx_hash)

    @staticmethod
    def get_deploy_info(context: 'IconScoreContext', address: 'Address') -> Optional['IconScoreDeployInfo']:
        return context.storage.deploy.get_deploy_info(context, address)
